# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2011, 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Flask :class:`~flask.sessions.SessionInterface` implementation."""

from datetime import datetime, timedelta
from uuid import uuid4

import six
from flask import current_app, request
from flask.helpers import locked_cached_property
from flask.sessions import SessionInterface as FlaskSessionInterface
from itsdangerous import BadSignature, TimestampSigner
from werkzeug.exceptions import BadRequest
from werkzeug.utils import import_string
from werkzeug.local import LocalProxy

from invenio_utils.serializers import LzmaPickle as Serializer

secret_key = LocalProxy(
    lambda: current_app.config['SECRET_KEY'].encode('utf-8')
)


class SessionInterface(FlaskSessionInterface):

    """Extend :class:`~flask.sessions.SessionInterface` class."""

    @locked_cached_property
    def has_secure_url(self):
        """Return ``True`` if secure url is configured."""
        return current_app.config.get('CFG_SITE_SECURE_URL', '').\
            startswith("https://")

    @locked_cached_property
    def serializer(self):
        """Return serializer class."""
        serializer_string = current_app.config.get('SESSION_SERIALIZER',
                                                   Serializer)
        return import_string(serializer_string)() \
            if isinstance(serializer_string, six.string_types) \
            else serializer_string()

    @locked_cached_property
    def session_class(self):
        """Return session class."""
        session_class_string = current_app.config.get(
            'SESSION_CLASS', 'invenio_ext.session.legacy_session:Session')
        return import_string(session_class_string) \
            if isinstance(session_class_string, six.string_types) \
            else session_class_string

    @locked_cached_property
    def backend(self):
        """Return session backend."""
        storage_string = current_app.config.get(
            'SESSION_BACKEND', 'invenio_ext.session.backends.cache:Storage')
        return import_string(storage_string)() \
            if isinstance(storage_string, six.string_types) \
            else storage_string()

    @locked_cached_property
    def sign_sid(self):
        """Return ``True`` if cookie session id is signed."""
        return current_app.config.get('SESSION_SIGN', True)

    @locked_cached_property
    def sign_salt(self):
        """Return salt for session id signing."""
        return current_app.config.get('SESSION_SALT', 'inveniosalt')

    def get_signer(self):
        """Return signer for cookie content signing."""
        return TimestampSigner(secret_key, self.sign_salt)

    def generate_sid(self):
        """Generate unique session identifier."""
        sid = uuid4().hex
        return sid

    def get_session_expiration_time(self, app, session):
        """Return session expiration time."""
        if session.permanent:
            return app.permanent_session_lifetime
        return timedelta(days=1)

    def open_session(self, app, request):
        """Return session instance."""
        sid_signed = request.cookies.get(app.session_cookie_name) or \
            request.args.get('session_id')

        # extract sid from signed version
        sid = None
        if sid_signed:
            if self.sign_sid:
                s = self.get_signer()

                # get maximum age without asking the cache (for
                # session.permanent) so we need to assume the worst case
                max_age = max(
                    app.permanent_session_lifetime.total_seconds(),
                    timedelta(days=1).total_seconds()
                )
                try:
                    sid = s.unsign(
                        sid_signed,
                        max_age=max_age
                    )
                except BadSignature:
                    # just ignore hackers and give them a new sid
                    # (this might also occur in case of clock divergence, but
                    # that should be a rare case)
                    sid = None
            else:
                # we did not sign, so it is just the same
                sid = sid_signed

        if not sid:
            sid = self.generate_sid()
            return self.session_class(sid=sid)
        try:
            data = self.backend.get(sid)
            if data:
                session = self.session_class(self.serializer.loads(data),
                                             sid=sid)
                return session
        except Exception:
            # this might have multiple reasons:
            #   - backend error (cache object timeout, cache flushed, ...)
            #   - error during deserialization
            #   - cache data does not match session_class (upgrades?)
            #   - earth stopped rotating
            current_app.logger.warning(
                "Load session error. Returning empty session.",
                exc_info=True)
        return self.session_class(sid=sid)

    def save_session(self, app, session, response):
        """Save current session."""
        domain = self.get_cookie_domain(app)
        if not session:
            current_app.logger.debug("Empty session: " + str(request.url))
            return

        timeout = self.get_session_expiration_time(app, session)
        session_expiry = datetime.utcnow() + timeout
        max_age = cookie_expiry = None
        uid = session.uid
        if uid > -1 and session.permanent:
            max_age = app.permanent_session_lifetime
            cookie_expiry = session_expiry
        sid = session.sid

        # Set all user id keys for compatibility.
        if len(session.keys()) == 1 and '_id' in session:
            session.delete()
            return
        elif not session.modified:
            return

        session.uid = uid
        self.backend.set(sid,
                         self.serializer.dumps(dict(session)),
                         timeout=timeout)

        # sign sid
        if self.sign_sid:
            s = self.get_signer()
            sid_signed = s.sign(sid)
        else:
            # we did not sign, so it is just the same
            sid_signed = sid

        if not self.has_secure_url:
            response.set_cookie(app.session_cookie_name, sid_signed,
                                expires=cookie_expiry, httponly=True,
                                domain=domain, max_age=max_age)
        elif session.uid > 0:
            # User is authenticated, we shall use HTTPS then
            if request.scheme == 'https':
                response.set_cookie(app.session_cookie_name, sid_signed,
                                    expires=cookie_expiry, httponly=True,
                                    domain=domain, secure=True,
                                    max_age=max_age)
                response.set_cookie(app.session_cookie_name + 'stub', 'HTTPS',
                                    expires=cookie_expiry, httponly=True,
                                    domain=domain, max_age=max_age)
            else:
                raise BadRequest("The user is being authenticated over HTTP "
                                 "rather than HTTPS?")
        else:
            response.set_cookie(app.session_cookie_name, sid_signed, httponly=True,
                                domain=domain)
            response.set_cookie(app.session_cookie_name + 'stub', 'NO',
                                httponly=True, domain=domain)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Test *ArXiv* integration."""

from __future__ import absolute_import

import os

import httpretty

from invenio_ext.arxiv import Arxiv
from invenio_testing import InvenioTestCase

current_dir = os.path.dirname(__file__)


def load_data(filename):
    with open(os.path.join(current_dir, 'data', filename), 'r') as f:
        return f.read()


class ArxivMixin(InvenioTestCase):

    """Custom Arxiv configuration."""

    @property
    def config(self):
        """Remove Arxiv from extensions to get full control of the test."""
        from invenio_base.config import EXTENSIONS
        cfg = super(ArxivMixin, self).config
        cfg["EXTENSIONS"] = filter(
            lambda k: not k.startswith("invenio_ext.arxiv"),
            EXTENSIONS)
        cfg["ARXIV_API_URL"] = "http://export.example.org/oai2"
        cfg["CACHE_TYPE"] = "simple"
        return cfg


class TestArxiv(ArxivMixin):

    """Test of extension creation."""

    def test_creation(self):
        assert "arxiv" not in self.app.extensions
        Arxiv(app=self.app)
        assert isinstance(self.app.extensions["arxiv"], Arxiv)

    def test_creation_old_flask(self):
        # Simulate old Flask (pre 0.9)
        del self.app.extensions
        Arxiv(app=self.app)
        assert isinstance(self.app.extensions["arxiv"], Arxiv)

    def test_creation_init(self):
        assert "arxiv" not in self.app.extensions
        r = Arxiv()
        r.init_app(app=self.app)
        assert isinstance(self.app.extensions["arxiv"], Arxiv)

    def test_double_creation(self):
        Arxiv(app=self.app)
        self.assertRaises(RuntimeError, Arxiv, app=self.app)


class TestArxivQuery(ArxivMixin):

    """Test Arxiv query response parsing."""

    def setUp(self):
        self.arxiv = Arxiv(app=self.app)

    def tearDown(self):
        del self.arxiv

    @httpretty.activate
    def test_found_result(self):
        httpretty.register_uri(
            httpretty.GET,
            self.app.config["ARXIV_API_URL"] +
            "?verb=GetRecord&metadataPrefix=arXiv&identifier=oai:arXiv.org:1007.5048",
            body=load_data("response_export_arxiv.xml"),
            status=200
        )
        httpretty.register_uri(
            httpretty.GET,
            "http://" + self.app.config.get("ES_HOSTS", ['localhost'])[0] + ":9200/*",
            status=400
        )

        response = self.app.extensions["arxiv"].search("1007.5048")
        self.assertEqual(response.status_code, 200)

    @httpretty.activate
    def test_zero_results_found(self):
        httpretty.register_uri(
            httpretty.GET,
            self.app.config["ARXIV_API_URL"] +
            "?verb=GetRecord&metadataPrefix=arXiv&identifier=oai:arXiv.org:dead.beef",
            body=load_data("response_export_arxiv_zero.xml"),
            status=200
        )
        httpretty.register_uri(
            httpretty.GET,
            "http://" + self.app.config.get("ES_HOSTS", ['localhost'])[0] + ":9200/*",
            status=400
        )

        response = self.app.extensions["arxiv"].search("9999.9999")
        self.assertEqual(response.status_code, 404)

    @httpretty.activate
    def test_unsupported_versioning(self):
        httpretty.register_uri(
            httpretty.GET,
            self.app.config["ARXIV_API_URL"] +
            "?verb=GetRecord&metadataPrefix=arXiv&identifier=oai:arXiv.org:1007.5048v1",
            body=load_data("response_export_arxiv_versioning.xml"),
            status=200
        )
        httpretty.register_uri(
            httpretty.GET,
            "http://" + self.app.config.get("ES_HOSTS", ['localhost'])[0] + ":9200/*",
            status=400
        )

        response = self.app.extensions["arxiv"].search("1007.5048v1")
        self.assertEqual(response.status_code, 415)

    @httpretty.activate
    def test_malformed_arxiv_id(self):
        httpretty.register_uri(
            httpretty.GET,
            self.app.config["ARXIV_API_URL"] +
            "?verb=GetRecord&metadataPrefix=arXiv&identifier=oai:arXiv.org:dead.beef",
            body=load_data("response_export_arxiv_malformed.xml"),
            status=200
        )
        httpretty.register_uri(
            httpretty.GET,
            "http://" + self.app.config.get("ES_HOSTS", ['localhost'])[0] + ":9200/*",
            status=400
        )

        response = self.app.extensions["arxiv"].search("dead.beef")
        self.assertEqual(response.status_code, 422)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2011, 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""SQLAlchemy backend for session.

Configuration variables for SQLAlchemy backend.

=================================== ===========================================
`SESSION_BACKEND_SQLALCHEMY`        Configured *Flask-SQLAlchemy* object.
                                    **Default:** ``invenio_ext.sqlalchemy:db``
`SESSION_BACKEND_SQLALCHEMY_MODEL`  SQLAlchemy ORM model. **Default:**
                                    ``invenio_ext.session.model:Session'``
`SESSION_BACKEND_SQLALCHEMY_GETTER` Name of method on model to retrieve session
                                    data. **Default:** ``get_session``
`SESSION_BACKEND_SQLALCHEMY_SETTER` Name of method on medel to set session
                                    data. **Default:** ``set_session``
`SESSION_BACKEND_SQLALCHEMY_VALUE`  Name of model attribute where to store
                                    session data. **Default:**
                                    ``session_object``
=================================== ===========================================
"""

from flask import current_app
from flask.helpers import locked_cached_property
from werkzeug import import_string

from ..storage import SessionStorage


class Storage(SessionStorage):

    """Implement database backend for SQLAlchemy model storage."""

    def __init__(self, *args, **kwargs):
        """Initialize database backend and create table if necessary."""
        if not self.db.engine.dialect.has_table(self.db.engine,
                                                self.model.__tablename__):
            self.model.__table__.create(bind=self.db.engine)
            self.db.session.commit()

    @locked_cached_property
    def db(self):
        """Return SQLAlchemy database object."""
        return import_string(current_app.config.get(
            'SESSION_BACKEND_SQLALCHEMY', 'invenio_ext.sqlalchemy:db'))

    @locked_cached_property
    def model(self):
        """Return SQLAlchemy model."""
        return import_string(current_app.config.get(
            'SESSION_BACKEND_SQLALCHEMY_MODEL',
            'invenio_ext.session.model:Session'))()

    @locked_cached_property
    def getter(self):
        """Return method to get session value."""
        return getattr(self.model, current_app.config.get(
            'SESSION_BACKEND_SQLALCHEMY_GETTER', 'get_session'
        ))

    @locked_cached_property
    def setter(self):
        """Return method to set session value."""
        return getattr(self.model, current_app.config.get(
            'SESSION_BACKEND_SQLALCHEMY_SETTER', 'set_session'
        ))

    @locked_cached_property
    def value(self):
        """Return model property for storing session value."""
        return current_app.config.get('SESSION_BACKEND_SQLALCHEMY_VALUE',
                                      'session_object')

    def set(self, name, value, timeout=None):
        """Store value in database table."""
        s = self.setter(name, value, timeout=timeout)
        self.db.session.merge(s)
        self.db.session.commit()

    def get(self, name):
        """Return value from database table."""
        return getattr(self.getter(name), self.value)

    def delete(self, name):
        """Delete key from database table."""
        self.getter(name).delete()
        self.db.session.commit()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Legacy Invenio hash support."""

import hashlib

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from passlib.utils.compat import str_to_uascii
from passlib.utils.handlers import HasUserContext, StaticHandler
from six import binary_type, text_type

__all__ = ('invenio_aes_encrypted_email', )


def _to_binary(val):
    if isinstance(val, text_type):
        return val.encode('utf-8')
    return val


def _mysql_aes_key(key):
    final_key = bytearray(16)
    for i, c in enumerate(key):
        final_key[i % 16] ^= ord(key[i])
    return bytes(final_key)


def _mysql_aes_pad(val):
    pad_value = 16 - (len(val) % 16)
    return '%s%s' % (val, chr(pad_value) * pad_value)


def _mysql_aes_unpad(val):
    pad_value = ord(val[-1])
    return val[:-pad_value]


def _mysql_aes_engine(key):
    """Create MYSQL AES cipher engine."""
    return Cipher(algorithms.AES(key), modes.ECB(), default_backend())


def mysql_aes_encrypt(val, key):
    """MySQL AES encrypt value with secret key."""
    assert isinstance(val, binary_type) or isinstance(val, text_type)
    assert isinstance(key, binary_type) or isinstance(key, text_type)
    k = _mysql_aes_key(_to_binary(key))
    v = _mysql_aes_pad(_to_binary(val))
    e = _mysql_aes_engine(k).encryptor()

    return e.update(v) + e.finalize()


def mysql_aes_decrypt(encrypted_val, key):
    """MySQL AES decrypt value with secret key."""
    assert isinstance(encrypted_val, binary_type) \
        or isinstance(encrypted_val, text_type)
    assert isinstance(key, binary_type) or isinstance(key, text_type)
    k = _mysql_aes_key(_to_binary(key))
    d = _mysql_aes_engine(_to_binary(k)).decryptor()
    return _mysql_aes_unpad(d.update(_to_binary(encrypted_val)) + d.finalize())


class invenio_aes_encrypted_email(HasUserContext, StaticHandler):

    """Invenio AES encryption of user email using password as secret key.

    Invenio 1.x was AES encrypting the users email address with the password
    as the secret key and storing it in a blob column. This e.g. caused
    problems when a user wanted to change email address.

    This hashing engine, differs from Invenio 1.x in that it sha256 hashes the
    encrypted value as well to produce a string in the same length instead of
    a binary blob. It is not done for extra security, just for convenience of
    migration to using passlib's sha512.

    An upgrade recipe is provided to migrated existing binary password hashes
    to hashes of this engine.
    """

    name = "invenio_aes_encrypted_email"

    def _calc_checksum(self, secret):
        """Calculate string."""
        return str_to_uascii(
            hashlib.sha256(mysql_aes_encrypt(self.user, secret)).hexdigest()
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Flask-IIIF extension.

.. py:data:: IIIF_IMAGE_OPENER

    There are two ways to initialize an IIIF image object, by ``fullpath`` and
    ``bytestream``. Can be an ``import string`` or a ``callback function``.
    By default ``identifier_to_path`` return the ``fullpath`` of
    :class:`~invenio_documents.api:Document`.

    default: :func:`~invenio_documents.utis.identifier_to_path`
"""

from flask_iiif import IIIF
from flask_iiif.errors import MultimediaError

from six import string_types

from werkzeug.utils import import_string

from .utils import api_file_permission_check

__all__ = ('setup_app', )

iiif = IIIF()


def setup_app(app):
    """Setup Flask-IIIF extension."""
    if 'invenio_documents' in app.config.get('PACKAGES_EXCLUDE'):
        raise MultimediaError(
            "Could not initialize the Flask-IIIF extension because "
            ":class:`~invenio_documents.api:Document` is missing"
        )

    iiif.init_app(app)
    iiif.init_restful(app.extensions['restful'])
    app.config.setdefault(
        'IIIF_IMAGE_OPENER',
        'invenio_documents.utils:identifier_to_path'
    )

    uuid_to_source_handler = app.config['IIIF_IMAGE_OPENER']

    uuid_to_source = (
        import_string(uuid_to_source_handler) if
        isinstance(uuid_to_source_handler, string_types) else
        uuid_to_source_handler
    )
    iiif.uuid_to_image_opener_handler(uuid_to_source)
    app.config['IIIF_CACHE_HANDLER'] = 'invenio_ext.cache:cache'

    # protect the api
    iiif.api_decorator_handler(api_file_permission_check)
    return app



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Bundles for Jasmine test runner."""

from __future__ import unicode_literals

from invenio_base.bundles import invenio as _i
from invenio_base.bundles import jquery as _j
from invenio_ext.assets import Bundle, RequireJSFilter

jasmine_js = Bundle(
    # es5-shim is needed by PhantomJS
    # 'vendors/es5-shim/es5-shim.js',
    # 'vendors/es5-shim/es5-sham.js',
    "js/jasmine/init.js",
    output="jasmine.js",
    weight=50,
    filters=RequireJSFilter(exclude=[_j, _i]),
    bower={
        "jasmine": ">=2",
        "jasmine-jquery": ">=2",
        "jasmine-flight": ">=3",
        "jasmine-ajax": ">=2",
    }
)

jasmine_styles = Bundle(
    'vendors/jasmine/lib/jasmine-core/jasmine.css',
    weight=-1,
    output='jasmine.css'
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2013,
#               2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Error handling library."""

from __future__ import absolute_import, print_function

import inspect
import os
import re
import sys
import time
import traceback
import warnings

from flask import current_app
from flask_login import current_user as user_info
from six import StringIO, iteritems

from invenio_base.globals import cfg

# Regular expression to match possible password related variable that should
# be disclosed in frame analysis.
RE_PWD = re.compile(r"pwd|pass|p_pw", re.I)


def get_pretty_wide_client_info(req):
    """Return (in a pretty way) all the available information about the current
    user/client"""
    if req:
        keys = sorted(user_info.keys())
        max_key = max([len(key) for key in keys])
        ret = ""
        fmt = "%% %is: %%s\n" % max_key
        for key in keys:
            if RE_PWD.search(key):
                continue
            if key in ('uri', 'referer'):
                ret += fmt % (key, "<%s>" % user_info[key])
            else:
                ret += fmt % (key, user_info[key])
        if ret.endswith('\n'):
            return ret[:-1]
        else:
            return ret
    else:
        return "No client information available"


def get_traceback():
    """
    If an exception has been caught, return the system stack trace or else
    return stack trace of what is currently in the stack
    """
    if traceback.format_tb(sys.exc_info()[2]):
        delimiter = "\n"
        traceback_pretty = "Traceback: \n%s" % \
            delimiter.join(traceback.format_tb(sys.exc_info()[2]))
    else:
        # force traceback except for this call
        stack = traceback.extract_stack()[:-1]
        traceback_pretty = "%sForced traceback (most recent call last)" % \
            (' ' * 4, )
        for trace_tuple in stack:
            traceback_pretty += """
    File "%(file)s", line %(line)s, in %(function)s
        %(text)s""" % {
                'file': trace_tuple[0],
                'line': trace_tuple[1],
                'function': trace_tuple[2],
                'text': trace_tuple[3] is not None
                and str(trace_tuple[3]) or ""
            }
    return traceback_pretty


def find_all_values_to_hide(local_variables, analyzed_stack=None):
    """Return all the potential password to hide."""
    # Let's add at least the DB password.
    if analyzed_stack is None:
        ret = set([cfg['CFG_DATABASE_PASS']])
        analyzed_stack = set()
    else:
        ret = set()
    for key, value in iteritems(local_variables):
        if id(value) in analyzed_stack:
            # Let's avoid loops
            continue
        analyzed_stack.add(id(value))
        if RE_PWD.search(key):
            ret.add(str(value))
        if isinstance(value, dict):
            ret |= find_all_values_to_hide(value, analyzed_stack)
    if '' in ret:
        # Let's discard the empty string in case there is an empty password,
        # or otherwise anything will be separated by '<*****>' in the output
        # :-)
        ret.remove('')
    return ret


def get_pretty_traceback(req=None, exc_info=None, skip_frames=0):
    """
    Given an optional request object and an optional exc_info,
    returns a text string representing many details about an exception.
    """
    if exc_info is None:
        exc_info = sys.exc_info()
    if exc_info[0]:
        # We found an exception.

        # We want to extract the name of the Exception
        exc_name = exc_info[0].__name__
        exc_value = str(exc_info[1])
        filename, line_no, function_name = _get_filename_and_line(exc_info)

        # Let's record when and where and what
        www_data = (
            "%(time)s -> %(name)s: %(value)s (%(file)s:%(line)s:%(function)s)"
            % {
                'time': time.strftime("%Y-%m-%d %H:%M:%S"),
                'name': exc_name,
                'value': exc_value,
                'file': filename,
                'line': line_no,
                'function': function_name
            }
        )

        # Let's retrieve contextual user related info, if any
        try:
            client_data = get_pretty_wide_client_info(req)
        except Exception as err:
            client_data = "Error in retrieving " \
                "contextual information: %s" % err

        # Let's extract the traceback:
        traceback_data_stream = StringIO()
        print("\n** Traceback details \n", file=traceback_data_stream)
        traceback.print_exc(file=traceback_data_stream)
        stack = [frame[0] for frame in inspect.trace()]
        try:
            stack.reverse()
            print("\n** Stack frame details", file=traceback_data_stream)
            values_to_hide = set()
            for frame in stack:
                try:
                    print(file=traceback_data_stream)
                    print("Frame %s in %s at line %s" % (
                        frame.f_code.co_name,
                        frame.f_code.co_filename,
                        frame.f_lineno), file=traceback_data_stream)
                    # Dereferencing f_locals
                    # See: http://utcc.utoronto.ca/~cks/space/blog/python/
                    # FLocalsAndTraceFunctions
                    local_values = frame.f_locals
                    try:
                        values_to_hide |= find_all_values_to_hide(local_values)

                        code = open(frame.f_code.co_filename).readlines()
                        first_line = max(1, frame.f_lineno - 3)
                        last_line = min(len(code), frame.f_lineno + 3)
                        print("-" * 79, file=traceback_data_stream)
                        for line in xrange(first_line, last_line + 1):
                            code_line = code[line - 1].rstrip()
                            if line == frame.f_lineno:
                                print(
                                    "----> %4i %s" % (line, code_line),
                                    file=traceback_data_stream
                                )
                            else:
                                print(
                                    "      %4i %s" % (line, code_line),
                                    file=traceback_data_stream
                                )
                        print("-" * 79, file=traceback_data_stream)
                    except:
                        pass
                    for key, value in local_values.items():
                        print(
                            "\t%20s = " % key, end=' ',
                            file=traceback_data_stream
                        )
                        try:
                            value = repr(value)
                        except Exception as err:
                            # We shall gracefully accept errors when repr() of
                            # a value fails (e.g. when we are trying to repr()
                            # a variable that was not fully initialized as the
                            # exception was raised during its __init__ call).
                            value = "ERROR: when representing the value: %s" \
                                    % (err)
                        try:
                            print(
                                _truncate_dynamic_string(value),
                                file=traceback_data_stream
                            )
                        except:
                            print(
                                "<ERROR WHILE PRINTING VALUE>",
                                file=traceback_data_stream
                            )
                finally:
                    del frame
        finally:
            del stack
        traceback_data = traceback_data_stream.getvalue()
        for to_hide in values_to_hide:
            # Let's hide passwords
            traceback_data = traceback_data.replace(to_hide, '<*****>')

        # Okay, start printing:
        output = StringIO()

        print("* %s" % www_data, file=output)
        print("\n** User details", file=output)
        print(client_data, file=output)

        if traceback_data:
            print(traceback_data, file=output)
        return output.getvalue()
    else:
        return ""


def register_exception(stream='error',
                       req=None,
                       prefix='',
                       suffix='',
                       alert_admin=False,
                       subject=''):
    """
    Log error exception to invenio.err and warning exception to invenio.log.
    Errors will be logged together with client information (if req is
    given).

    Note:   For sanity reasons, dynamic params such as PREFIX, SUFFIX and
            local stack variables are checked for length, and only first 500
            chars of their values are printed.

    :param stream: 'critical', 'error', 'warning', 'info', 'debug'

    :param req: mod_python request

    :param prefix: a message to be printed before the exception in
    the log

    :param suffix: a message to be printed before the exception in
    the log

    :param alert_admin: whether to send the exception to the administrator via
        email. Note this parameter is bypassed when
        CFG_SITE_ADMIN_EMAIL_EXCEPTIONS is set to a value different than 1
    :param subject: overrides the email subject

    :return: 1 if successfully wrote to stream, 0 if not
    """
    if stream == 'error':
        logger = current_app.logger.error
    elif stream == 'warning':
        logger = current_app.logger.warning
    elif stream == 'info':
        logger = current_app.logger.info
    elif stream == 'debug':
        logger = current_app.logger.debug
    elif stream == 'critical':
        logger = current_app.logger.critical
    else:
        # Legacy - Invenio 1.x defaults to log level warning when stream was
        # undefined.
        warnings.warn(
            "register_exception() called without 'stream' defined",
            DeprecationWarning
        )
        logger = current_app.logger.warning

    extra = dict(
        invenio_register_exception=dict(
            stream=stream,
            req=req,
            prefix=prefix,
            suffix=suffix,
            alert_admin=alert_admin,
            subject=subject,
        )
    )

    try:
        logger(
            subject,
            exc_info=sys.exc_info(),
            extra=extra
        )
        return 1
    except Exception as err:
        print("Error in registering exception '%s'" % err, file=sys.stderr)
        return 0


def _get_filename_and_line(exc_info):
    """Return the filename, the line and the function_name where
    the exception happened."""
    tb = exc_info[2]
    exception_info = traceback.extract_tb(tb)[-1]
    filename = os.path.basename(exception_info[0])
    line_no = exception_info[1]
    function_name = exception_info[2]
    return filename, line_no, function_name


def _truncate_dynamic_string(val, maxlength=500):
    """
    Return at most MAXLENGTH characters of VAL.  Useful for
    sanitizing dynamic variable values in the output.
    """
    out = repr(val)
    if len(out) > maxlength:
        out = out[:maxlength] + ' [...]'
    return out



# -*- coding: utf-8 -*-

# This file is part of Invenio.
# Copyright (C) 2011, 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Implementation of legacy Invenio methods for Flask session."""

from flask import current_app, request
from flask.sessions import SessionMixin
from flask_login import current_user
from werkzeug.datastructures import CallbackDict


class Session(CallbackDict, SessionMixin):

    """Implement compatible legacy Invenio session."""

    def __init__(self, initial=None, sid=None):
        """Initialize session with optional default value."""
        self.sid = sid
        self.logging_in = False
        self.modified = initial is not None

        def _on_update(d):
            d.modified = True

        CallbackDict.__init__(self, initial, _on_update)

    def need_https(self):
        """Check if the user was previously authenticated.

        If True session identifier need to be sent via HTTPS.
        """
        return request.cookies.get(
            current_app.session_cookie_name + 'stub', 'NO') == 'HTTPS'

    def delete(self, clear=True):
        """Delete the session."""
        if clear:
            self.clear()

    def invalidate(self):
        """Declare the session as invalid."""
        self._invalid = 1

    def set_remember_me(self, remember_me=True):
        """Set or unset the ``_remember_me`` flag.

        :param remember_me: True if the session cookie should last one day or
            until the browser is closed.
        """
        self._remember_me = remember_me
        self['_permanent'] = remember_me

    def _get_uid(self):
        return self.get('user_id', -1)

    def _set_uid(self, uid):
        if self.get('user_id') != uid:
            self.logging_in = True
        self['user_id'] = self['_uid'] = self['uid'] = uid

    def _get_user_info(self):
        return current_user

    uid = property(_get_uid, _set_uid)
    user_info = property(_get_user_info)
    del _get_uid, _set_uid, _get_user_info



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2011, 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Mysql dialect."""

import base64

import sqlalchemy
from sqlalchemy import types as types
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.types import TypeDecorator


@compiles(types.Text, 'sqlite')
@compiles(sqlalchemy.dialects.mysql.TEXT, 'sqlite')
def compile_text(element, compiler, **kw):
    """Redefine Text filed type for SQLite and MySQL."""
    return 'TEXT'


@compiles(types.Binary, 'sqlite')
def compile_binary(element, compiler, **kw):
    """Redefine Binary filed type for SQLite."""
    return 'BLOB'


@compiles(types.LargeBinary, 'sqlite')
def compile_largebinary(element, compiler, **kw):
    """Redefine LargeBinary filed type for SQLite."""
    return 'LONGBLOB'


@compiles(types.Text, 'mysql')
@compiles(sqlalchemy.dialects.mysql.TEXT, 'mysql')
def compile_text(element, compiler, **kw):
    """Redefine Text filed type for MySQL."""
    return 'TEXT'


@compiles(types.Binary, 'mysql')
def compile_binary(element, compiler, **kw):
    """Redefine Binary filed type for MySQL."""
    return 'BLOB'


@compiles(types.LargeBinary, 'mysql')
def compile_largebinary(element, compiler, **kw):
    """Redefine LargeBinary filed type for MySQL."""
    return 'LONGBLOB'


class iBinary(TypeDecorator):

    """Printable binary typea."""

    impl = types.Binary

    def __init__(self, *arg, **kw):
        """Init iBinary type."""
        self.__class__.impl = self.impl
        TypeDecorator.__init__(self, *arg, **kw)

    def process_bind_param(self, value, dialect):
        """Decode string before saving to database."""
        return (value is not None) and base64.decodestring(value) or None

    def process_result_value(self, value, dialect):
        """Encode binary data to string."""
        return (value is not None) and base64.encodestring(value) or None


class iLargeBinary(TypeDecorator):

    """Printable large binary type."""

    impl = types.LargeBinary

    def __init__(self, *arg, **kw):
        """Init iLargeBinary type."""
        self.__class__.impl = self.impl
        TypeDecorator.__init__(self, *arg, **kw)

    def process_bind_param(self, value, dialect):
        """Decode string before saving to database."""
        return (value is not None) and base64.decodestring(value) or None

    def process_result_value(self, value, dialect):
        """Encode binary data to string."""
        return (value is not None) and base64.encodestring(value) or None


class iMediumBinary(TypeDecorator):

    """Printable large binary type."""

    impl = sqlalchemy.dialects.mysql.MEDIUMBLOB

    def __init__(self, *arg, **kw):
        """Init iMediumBinary type."""
        self.__class__.impl = self.impl
        TypeDecorator.__init__(self, *arg, **kw)

    def process_bind_param(self, value, dialect):
        """Decode string before saving to database."""
        return (value is not None) and base64.decodestring(value) or None

    def process_result_value(self, value, dialect):
        """Encode binary data to string."""
        return (value is not None) and base64.encodestring(value) or None



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Implement various localization selectors."""

from flask import current_app, request, session
from flask_login import current_user


def get_locale():
    """Compute the language needed to return the answer to the client."""
    from invenio_base.i18n import wash_language
    required_ln = None
    passed_ln = request.values.get('ln', type=str)
    if passed_ln:
        # If ln is specified explictly as a GET or POST argument
        # let's take it!
        required_ln = wash_language(passed_ln)
        if passed_ln != required_ln:
            # But only if it was a valid language
            required_ln = None
    if required_ln is None and 'ln' not in session:
        # If there is no language saved into the session...
        user_language = current_user.get('language')
        if user_language:
            # ... and the user is logged in, we try to take it from its
            # settings.
            required_ln = user_language
        else:
            # Otherwise we try to guess it from its request headers
            for value, quality in request.accept_languages:
                value = str(value)
                ln = wash_language(value)
                if ln == value or ln[:2] == value[:2]:
                    required_ln = ln
                    break
            else:
                # Too bad! We stick to the default :-)
                required_ln = current_app.config.get('CFG_SITE_LANG')
    elif required_ln is None:
        required_ln = session.get('ln')

    assert required_ln is not None

    if required_ln != session.get('ln',
                                  current_app.config.get('CFG_SITE_LANG')):
        session['ln'] = required_ln

    return required_ln


def get_timezone():
    """Return the timezone from user settings."""
    return current_user.get('timezone')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Sentry logging backend.

Currently only Python application errors are sent to
`Sentry <http://getsentry.com>`_. Future extensions may allow for sending
JavaScript errors to Sentry as well.

**Configuration**

================================== ============================================
`SENTRY_DSN`                       Sentry DSN (get it from your Sentry account)
                                   . **Required**.
`LOGGING_SENTRY_LEVEL`             Log level threshold for handler.
                                   **Default:**  ``WARNING``.
`LOGGING_SENTRY_INCLUDE_WARNINGS`  Include messages from warnings module.
                                   **Default:**  ``True``.
`LOGGING_SENTRY_CELERY`            Log Celery messages to Sentry.
                                   **Default:**  ``True``.
`LOGGING_SENTRY_CELERY_TRANSPORT`  Transport mechanism for Celery.
                                   **Default:**  ``sync``.
================================== ============================================

`Raven <raven.readthedocs.org/en/latest/>`_ (the Python library responsible for
sending log messages to Sentry), supports some additionally configuration
variables. See https://github.com/getsentry/raven-python/blob/master/raven/contrib/flask.py
for further details.
"""

import logging
from functools import partial

import pkg_resources
from celery.signals import after_setup_logger, after_setup_task_logger
from flask import current_app
from raven.contrib.flask import Sentry
from raven.handlers.logging import SentryHandler
from raven.processors import SanitizePasswordsProcessor
from werkzeug.local import LocalProxy


class InvenioSanitizeProcessor(SanitizePasswordsProcessor):

    """Remove additional sensitve configuration from Sentry data."""

    FIELDS = frozenset([
        'access_token'
    ])


def sentry_include_paths():
    """Detect Invenio dependencies and for use with SENTRY_INCLUDE_PATHS."""
    try:
        dist = pkg_resources.get_distribution('invenio')
        return map(lambda req: req.key, dist.requires())
    except pkg_resources.DistributionNotFound:
        pass


def setup_warnings(sentry):
    """Add sentry to warnings logger."""
    warnings = logging.getLogger('py.warnings')
    warnings.addHandler(SentryHandler(sentry.client, level=logging.WARNING))


def add_sentry_id_header(self, sender, response, *args, **kwargs):
    """Fix issue when last_event_id is not defined."""
    if hasattr(self, 'last_event_id'):
        response.headers['X-Sentry-ID'] = self.last_event_id
    return response


def celery_logger_setup(app=None, sender=None, logger=None, **kwargs):
    """Setup Sentry logging for Celery."""
    add_handler(logger, app)


def celery_dsn_fix(app):
    """Fix SENTRY_DSN for Celery.

    Celery does not handle threaded transport very well, so allow overriding
    default transport mechanism for Celery.
    """
    if app.config.get('CELERY_CONTEXT', False) and \
       app.config['LOGGING_SENTRY_CELERY'] and \
       app.config['LOGGING_SENTRY_CELERY_TRANSPORT']:
        parts = app.config['SENTRY_DSN'].split('+', 1)
        if parts[0] in ('eventlet', 'gevent', 'requests', 'sync',
                        'threaded', 'twisted', 'tornado'):
            app.config['SENTRY_DSN'] = "%s+%s" % (
                app.config['LOGGING_SENTRY_CELERY_TRANSPORT'],
                parts[1],
            )
        else:
            app.config['SENTRY_DSN'] = "%s+%s" % (
                app.config['LOGGING_SENTRY_CELERY_TRANSPORT'],
                "+".join(parts),
            )


def add_handler(logger, app):
    """Add handler to logger if not already added."""
    for h in logger.handlers:
        if isinstance(h, SentryHandler):
            return

    logger.addHandler(
        SentryHandler(
            app.extensions['sentry'].client,
            level=app.config['LOGGING_SENTRY_LEVEL']
        )
    )


def setup_app(app):
    """Setup Sentry extension."""
    app.config.setdefault('SENTRY_DSN', None)
    # Sanitize data more
    app.config.setdefault('SENTRY_PROCESSORS', (
        'raven.processors.SanitizePasswordsProcessor',
        'invenio_ext.logging.backends.sentry.InvenioSanitizeProcessor',
    ))
    # When a user is logged in, also include the user info in the log message.
    app.config.setdefault('SENTRY_USER_ATTRS', ['info', ])
    # Defaults to only reporting errors and warnings.
    app.config.setdefault('LOGGING_SENTRY_LEVEL', 'WARNING')
    # Send warnings to Sentry?
    app.config.setdefault('LOGGING_SENTRY_INCLUDE_WARNINGS', True)
    # Send Celery log messages to Sentry?
    app.config.setdefault('LOGGING_SENTRY_CELERY', True)
    # Transport mechanism for Celery. Defaults to synchronous transport.
    # See http://raven.readthedocs.org/en/latest/transports/index.html
    app.config.setdefault('LOGGING_SENTRY_CELERY_TRANSPORT', 'sync')

    if app.config['SENTRY_DSN']:
        # Detect Invenio requirements and add to Sentry include paths so
        # version information about them is added to the log message.
        app.config.setdefault('SENTRY_INCLUDE_PATHS', sentry_include_paths())

        # Fix-up known version problems getting version information
        # Patch submitted to raven-python, if accepted the following lines
        # can be removed:
        # https://github.com/getsentry/raven-python/pull/452
        from raven.utils import _VERSION_CACHE
        import invenio
        import webassets
        import setuptools
        _VERSION_CACHE['invenio'] = invenio.__version__
        _VERSION_CACHE['webassets'] = webassets.__version__
        _VERSION_CACHE['setuptools'] = setuptools.__version__

        # Modify Sentry transport for Celery - must be called prior to client
        # creation.
        celery_dsn_fix(app)

        # Installs sentry in app.extensions['sentry']
        s = Sentry(
            app,
            logging=True,
            level=getattr(logging, app.config['LOGGING_SENTRY_LEVEL'])
        )

        # Replace method with more robust version
        s.add_sentry_id_header = add_sentry_id_header

        # Add extra tags information to sentry.
        s.client.extra_context({'version': invenio.__version__})

        # Capture warnings from warnings module
        if app.config['LOGGING_SENTRY_INCLUDE_WARNINGS']:
            setup_warnings(s)

        # Setup Celery logging to Sentry
        if app.config['LOGGING_SENTRY_CELERY']:
            # Setup Celery loggers
            after_setup_task_logger.connect(
                partial(celery_logger_setup, app=app),
                weak=False
            )
            after_setup_logger.connect(
                partial(celery_logger_setup, app=app),
                weak=False
            )

        # Werkzeug only adds a stream handler if there's no other handlers
        # defined, so when Sentry adds a log handler no output is
        # received from Werkzeug unless we install a console handler here on
        # the werkzeug logger.
        if app.debug:
            logger = logging.getLogger('werkzeug')
            logger.setLevel(logging.INFO)
            handler = logging.StreamHandler()
            logger.addHandler(handler)


sentry = LocalProxy(lambda: current_app.extensions['sentry'])
"""Proxy object to sentry instance."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.


from __future__ import absolute_import

import logging

from invenio_base.globals import cfg

from .models import HstEXCEPTION
from .wrappers import _get_filename_and_line


class InvenioLegacyHandler(logging.Handler):
    """
    Invenio log record handler. Use together with InvenioExceptionFormatter.
    """

    def emit(self, record):
        # Set by invenio_ext.logging.wrappers.register_exception
        if hasattr(record, 'invenio_register_exception'):
            extra = record.invenio_register_exception

            exc_info = record.exc_info
            exc_name = exc_info[0].__name__
            alert_admin = extra['alert_admin']
            subject = extra['subject']
            content = self.format(record)
            filename, line_no, function_name = _get_filename_and_line(exc_info)

            # let's log the exception and see whether we should report it.
            log = HstEXCEPTION.get_or_create(exc_name, filename, line_no)
            if log.exception_should_be_notified and (
               cfg['CFG_SITE_ADMIN_EMAIL_EXCEPTIONS'] > 1 or
               (alert_admin and cfg['CFG_SITE_ADMIN_EMAIL_EXCEPTIONS'] > 0)):

                # Set subject of email
                if not subject:
                    subject = 'Exception (%s:%s:%s)' % (
                        filename, line_no, function_name
                    )
                subject = '%s at %s' % (subject, cfg['CFG_SITE_URL'])

                # Set content of email
                content = "\n%s\n%s" % (
                    log.pretty_notification_info,
                    content
                )

                # Send the email
                from invenio_ext.email import send_email
                send_email(
                    cfg['CFG_SITE_ADMIN_EMAIL'],
                    cfg['CFG_SITE_ADMIN_EMAIL'],
                    subject=subject,
                    content=content
                )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Restful pagination."""

from flask import request, url_for

from invenio_utils.pagination import Pagination as InvenioPagination

from .errors import InvalidPageError


class RestfulPaginationMixIn(object):

    """Provide implementaion methods :meth:`link_header` and :meth:`links`."""

    def link_header(self, **kwargs):
        """Return tuple that can be used in response header."""
        links = self.links(**kwargs)
        keys = ("first", "prev", "next", "last")
        links_string = ",".join([links[key] for key in keys if key in links])
        return ('Link', links_string)

    def links(self, endpoint=None, args=None):
        """Generate links for the headers.

        :param endpoint: the URL endpoint
        :param args: the arguments of request
        :param count: the total number of items
        """
        if not endpoint:
            endpoint = request.endpoint
        if not args:
            args = request.args

        links = {}
        link_template = '<{}>; rel="{}"'

        # arguments to stick to the URL
        url_args = {}
        # url_args['page'] will be updated for every link
        url_args['page'] = 1
        url_args['per_page'] = self.per_page

        # generate link for rel first
        links['first'] = link_template.format(
            url_for(endpoint, **url_args), "first"
        )

        # generate link for prev if it exists
        if self.has_prev:
            url_args['page'] = self.page - 1
            links['prev'] = link_template.format(
                url_for(endpoint, **url_args), "prev"
            )

        # generate link for next if it exists
        if self.has_next:
            url_args['page'] = self.page + 1
            links['next'] = link_template.format(
                url_for(endpoint, **url_args), "next"
            )

        # generate link for last
        url_args['page'] = self.pages
        links['last'] = link_template.format(
            url_for(endpoint, **url_args), "last"
        )
        return links


class SQLAlchemyPaginationHelper(object):

    """Wrap SQLAlchemy query `paginate` method."""

    def __init__(self, query, page, per_page):
        """Initialize pagination property.

        :param query: query object from SQLAlchemy
        """
        self.query = query
        self.page = page
        self.per_page = per_page
        self.pagination = self.query.paginate(self.page, self.per_page, False)


class RestfulSQLAlchemyPagination(SQLAlchemyPaginationHelper,
                                  RestfulPaginationMixIn):

    """Implement Restful pagination for SQLAlchemy model."""

    @property
    def items(self):
        """Return found items."""
        return self.pagination.items

    @property
    def pages(self):
        """Return number of pages."""
        return self.pagination.pages

    @property
    def has_next(self):
        """Return `True` if has next page otherwise return `False`."""
        return self.pagination.has_next

    @property
    def has_prev(self):
        """Return `True` if has previous page otherwise return `False`."""
        return self.pagination.has_prev


class RestfulPagination(InvenioPagination, RestfulPaginationMixIn):

    """Implement Restful pagination for list of data."""

    def __init__(self, page, per_page, total_count, validate=True):
        super(RestfulPagination, self).__init__(page, per_page, total_count)
        if validate:
            self.validate()

    def validate(self):
        """Validate the range of page and per_page."""
        if self.per_page < 0:
            error_msg = (
                "Invalid per_page argument ('{0}'). Number of items "
                "per pages must be positive integer.".format(self.per_page)
            )
            raise InvalidPageError(error_msg)
        if self.page < 0 or self.page > self.pages:
            error_msg = "Invalid page number ('{0}').".format(self.page)
            raise InvalidPageError(error_msg)

    def slice(self, items):
        """Return items on current page."""
        return items[(self.page - 1) * self.per_page:self.page * self.per_page]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Extension for logging and registering errors.

In short, use the Flask application logger (``current_app.logger``). Do
not create log files manually.

To run the examples below in a shell, be sure to first create the Flask
application:

>>> from flask import current_app
>>> from invenio_base.factory import create_app
>>> app = create_app()

Logging errors
--------------
The preferred way to log errors is by using the Flask application logger:

>>> with app.app_context():
...     try:
...         raise Exception("This is an exception")
...     except Exception:
...         current_app.logger.exception("My message")

``logger.exception()`` will automatically include the exception stacktrace in
the log record, which each log handler may decide to include or not.

You may also manually include exception information in the logger using the
``exc_info`` keyword argument:

>>> import sys
>>> with app.app_context():
...     try:
...         raise Exception("This is an exception")
...     except Exception:
...         current_app.logger.critical("My message", exc_info=1)


Naturally, other log levels may also be used:

>>> app.logger.info("This is an info message")
>>> app.logger.warning("This is a warning message")
>>> app.logger.debug("This is a debug message")

Log handlers
------------
Log messages written to the Flask application logger can be handled by many
different backends, which is configurable by the an administrator. By default
Invenio ships with following log handlers:

* ``invenio_ext.logging.backends.fs`` - Rotating file system handler.
* ``invenio_ext.logging.backends.legacy`` - Error email reporting and logging
  to database. Default logging behaviour of Invenio 1.x.
* ``invenio_ext.logging.backends.sentry`` - Logging to Sentry service (see
  https://pypi.python.org/pypi/sentry and https://getsentry.com/)

Installing one or more of the logging backends is a simple as including
them in your configuration variable ``EXTENSIONS``::

    EXTENSIONS = [
        # ...
        'invenio_ext.logging',
        'invenio_ext.logging.backends.fs',
        'invenio_ext.logging.backends.legacy',
        'invenio_ext.logging.backends.sentry',
        # ...
    ]

Note that each backend may require additional configuration. Please see
:ref:`ext_logging_backends` for specific details.

Additionally if you plan to write your own backend, you may wish to consult
Python's logging documentation for how to create handlers, formatters
and filters: https://docs.python.org/2/library/logging.html

Legacy handling of errors
-------------------------
Invenio 1.x used a method ``register_exception`` to log errors. This method may
still be used, but may be deprecated in the future:

>>> from invenio_ext.logging import register_exception
>>> with app.app_context():
...     try:
...         raise Exception("This is an exception")
...     except Exception as e:
...         register_exception()


The method ``register_exception`` is in fact just a small wrapper around the
application logger.

Error handling do's and don'ts
------------------------------

**Always use ``except Exception:`` (or preferably more specific exceptions)
over ``except:``**, unless you explicitly want to catch the following built-in
exceptions (``SystemExit``, ``KeyboardInterrupt``, ``GeneratorExit``).

See https://docs.python.org/2/library/exceptions.html#exception-hierarchy

**Reraise**. To gracefully handle errors, you may often catch
exceptions to perform some cleanup or e.g. convert a low-level library
exception into a more high-level application exception. This may however often
discard the initial exception and its traceback, making it hard to track
down the root cause. To preserve the traceback, simply reraise the caught
exception using a ``raise`` with no arguments:

>>> with app.app_context():
...     try:
...         try:
...             0 / 0
...         except ZeroDivisionError as e:
...             # Do clean-up
...             raise
...     except Exception as e:
...         current_app.logger.exception("Something bad happened")

If you like to convert the exception, it can be done like this:

>>> import six
>>> class AppError(Exception):
...     pass
>>> with app.app_context():
...     try:
...         try:
...             0 / 0
...         except ZeroDivisionError as e:
...             six.reraise(MyAppError, "Custom message", sys.exc_info()[2])
...     except AppError as e:
...         current_app.logger.exception("Something bad happened")

Warnings
^^^^^^^^
Warnings are useful to alert developers and system administrators about
possible problems, e.g. usage of obsolete modules, deprecated APIs etc.

Please follow Invenio :ref:`deprecationpolicy` section.
"""

from __future__ import absolute_import

import logging

from .wrappers import get_pretty_traceback, register_exception


def setup_app(app):
    """Setup logging extesions."""
    # Output deprecation warnings in debug mode
    if app.debug:
        logger = logging.getLogger('py.warnings')
        logger.addHandler(logging.StreamHandler())
        logger.setLevel(logging.WARNING)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Dropbox file system.

Installation:: pip install invenio[dropbox]
"""

import calendar
import datetime
import os
import time
from UserDict import UserDict

from dropbox import client, rest
from fs.base import FS, NoDefaultMeta, synchronize
from fs.errors import CreateFailedError, DestinationExistsError, \
    DirectoryNotEmptyError, OperationFailedError, RemoteConnectionError, \
    ResourceInvalidError, ResourceNotFoundError, UnsupportedError
from fs.filelike import SpooledTemporaryFile
from fs.path import abspath, basename, dirname, normpath, pathsplit
from fs.remote import RemoteFileBuffer

# Items in cache are considered expired after 5 minutes.
CACHE_TTL = 300
# The format Dropbox uses for times.
TIME_FORMAT = '%a, %d %b %Y %H:%M:%S +0000'
# Max size for spooling to memory before using disk (5M).
MAX_BUFFER = 1024**2 * 5


class CacheItem(object):

    """Represent a path in the cache.

    There are two components to a path.
    It's individual metadata, and the children contained within it.
    """

    def __init__(self, metadata=None, children=None, timestamp=None):
        """Initialize a CacheItem instance."""
        self.metadata = metadata
        self.children = children
        if timestamp is None:
            timestamp = time.time()
        self.timestamp = timestamp

    def add_child(self, name, client=None):
        """Add a child."""
        if self.children is None:
            if client is not None:
                # This is a fix. When you add a child to a folder that
                # was still not listed, that folder gets only one
                # child when you list it afterwards. So this fix
                # first tries to check are the files/folders inside
                # this directory on cloud.
                client.children(self.metadata['path'])
            else:
                self.children = [name]
        else:
            if name not in self.children:
                self.children.append(name)

    def del_child(self, name):
        """Delete a child."""
        if self.children is None:
            return
        try:
            i = self.children.index(name)
        except ValueError:
            return
        self.children.pop(i)

    def _get_expired(self):
        if self.timestamp <= time.time() - CACHE_TTL:
            return True
    expired = property(_get_expired)

    def renew(self):
        """Renew."""
        self.timestamp = time.time()


class DropboxCache(UserDict):

    """Represent the dropbox cache."""

    def __init__(self, client):
        """Initialize a DropboxCache instance."""
        self._client = client
        UserDict.__init__(self)

    def set(self, path, metadata):
        """Set metadata."""
        self[path] = CacheItem(metadata)
        dname, bname = pathsplit(path)
        item = self.get(dname)
        if item:
            item.add_child(bname, self._client)

    def pop(self, path, default=None):
        """Pop data of a given path."""
        value = UserDict.pop(self, path, default)
        dname, bname = pathsplit(path)
        item = self.get(dname)
        if item:
            item.del_child(bname)
        return value


class DropboxClient(client.DropboxClient):

    """A wrapper around the official DropboxClient.

    This wrapper performs caching as well as converting
    errors to fs exceptions.
    """

    def __init__(self, *args, **kwargs):
        """Initialize a DropboxClient instance."""
        super(DropboxClient, self).__init__(*args, **kwargs)
        self.cache = DropboxCache(self)

    # Below we split the DropboxClient metadata() method into two methods
    # metadata() and children(). This allows for more fine-grained fetches
    # and caching.

    def metadata(self, path):
        """Get metadata for a given path."""
        item = self.cache.get(path)
        if not item or item.metadata is None or item.expired:
            try:
                metadata = super(
                    DropboxClient, self).metadata(
                    path, include_deleted=False, list=False)
            except rest.ErrorResponse as e:
                if e.status == 404:
                    raise ResourceNotFoundError(path)
                raise OperationFailedError(opname='metadata', path=path,
                                           msg=str(e))
            except:
                raise RemoteConnectionError(
                    "Most probable reasons: access token has expired or user"
                    " credentials are invalid.")
            if metadata.get('is_deleted', False):
                raise ResourceNotFoundError(path)
            item = self.cache[path] = CacheItem(metadata)
        # Copy the info so the caller cannot affect our cache.
        return dict(item.metadata.items())

    def children(self, path):
        """Get children of a given path."""
        update = False
        hash_ = None
        item = self.cache.get(path)
        if item:
            if item.expired:
                update = True
            if item.metadata and item.children:
                hash_ = item.metadata['hash']
            else:
                if not item.metadata.get('is_dir'):
                    raise ResourceInvalidError(path)
            if not item.children:
                update = True
        else:
            update = True
        if update:
            try:
                metadata = super(
                    DropboxClient, self).metadata(
                    path, hash=hash_, include_deleted=False, list=True)
                children = []
                contents = metadata.pop('contents')
                for child in contents:
                    if child.get('is_deleted', False):
                        continue
                    children.append(basename(child['path']))
                    self.cache[child['path']] = CacheItem(child)
                item = self.cache[path] = CacheItem(metadata, children)
            except rest.ErrorResponse as e:
                if not item or e.status != 304:
                    raise OperationFailedError(opname='metadata', path=path,
                                               msg=str(e))
                # We have an item from cache (perhaps expired), but it's
                # hash is still valid (as far as Dropbox is concerned),
                # so just renew it and keep using it.
                item.renew()
            except:
                raise RemoteConnectionError(
                    "Most probable reasons: access token has expired or user"
                    " credentials are invalid.")
        return item.children

    def file_create_folder(self, path):
        """Add newly created directory to cache."""
        try:
            metadata = super(DropboxClient, self).file_create_folder(path)
        except rest.ErrorResponse as e:
            if e.status == 403:
                raise DestinationExistsError(path)
            if e.status == 400:
                raise OperationFailedError(opname='file_create_folder',
                                           msg=str(e))
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired or user"
                " credentials are invalid.")
        self.cache.set(path, metadata)
        return metadata['path']

    def file_copy(self, src, dst):
        """Copy a file to another location."""
        try:
            metadata = super(DropboxClient, self).file_copy(src, dst)
        except rest.ErrorResponse as e:
            if e.status == 404:
                raise ResourceNotFoundError(src)
            if e.status == 403:
                raise DestinationExistsError(dst)
            if e.status == 503:
                raise OperationFailedError(opname='file_copy',
                                           msg="User over storage quota")
            raise OperationFailedError(opname='file_copy', msg=str(e))
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired or user"
                " credentials are invalid.")
        self.cache.set(dst, metadata)
        return metadata['path']

    def file_move(self, src, dst):
        """Move a file to another location."""
        try:
            metadata = super(DropboxClient, self).file_move(src, dst)
        except rest.ErrorResponse as e:
            if e.status == 404:
                raise ResourceNotFoundError(src)
            if e.status == 403:
                raise DestinationExistsError(dst)
            if e.status == 503:
                raise OperationFailedError(opname='file_copy',
                                           msg="User over storage quota")
            raise OperationFailedError(opname='file_copy', msg=str(e))
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired or user"
                " credentials are invalid.")
        self.cache.pop(src, None)
        self.cache.set(dst, metadata)
        return metadata['path']

    def file_delete(self, path):
        """Delete a file  of a give path."""
        try:
            super(DropboxClient, self).file_delete(path)
        except rest.ErrorResponse as e:
            if e.status == 404:
                raise ResourceNotFoundError(path)
            if e.status == 400 and 'must not be empty' in str(e):
                raise DirectoryNotEmptyError(path)
            raise OperationFailedError(opname='file_copy', msg=str(e))
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired or user"
                " credentials are invalid.")
        self.cache.pop(path, None)

    def put_file(self, path, f, overwrite=False):
        """Upload a file."""
        try:
            response = super(DropboxClient,
                             self).put_file(path, f, overwrite=overwrite)
        except rest.ErrorResponse as e:
            raise OperationFailedError(opname='file_copy', msg=str(e))
        except TypeError as e:
            raise ResourceInvalidError("put_file", path)
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired or user"
                " credentials are invalid.")
        self.cache.pop(dirname(path), None)
        return response

    def media(self, path):
        """Media."""
        try:
            info = super(DropboxClient, self).media(path)
            return info.get('url', None)
        except rest.ErrorResponse as e:
            if e.status == 400:
                raise UnsupportedError("create a link to a folder")
            if e.status == 404:
                raise ResourceNotFoundError(path)

            raise OperationFailedError(opname='file_copy', msg=str(e))
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired or user"
                " credentials are invalid.")


class DropboxFS(FS):

    """A Dropbox filesystem."""

    __name__ = "Dropbox"

    _meta = {'thread_safe': True,
             'virtual': False,
             'read_only': False,
             'unicode_paths': True,
             'case_insensitive_paths': True,
             'network': True,
             'atomic.setcontents': True,
             'atomic.makedir': True,
             'atomic.rename': True,
             'mime_type': 'virtual/dropbox',
             }

    def __init__(self, root=None, credentials=None, localtime=False,
                 thread_synchronize=True):
        """Initialize a DropboxFS instance."""
        self._root = root
        self._credentials = credentials

        if root is None:
            root = "/"

        if self._credentials is None:
            if "DROPBOX_ACCESS_TOKEN" not in os.environ:
                raise CreateFailedError(
                    "DROPBOX_ACCESS_TOKEN is not set in os.environ")
            else:
                self._credentials['access_token'] = os.environ.get(
                    'DROPBOX_ACCESS_TOKEN')

        super(DropboxFS, self).__init__(thread_synchronize=thread_synchronize)
        self.client = DropboxClient(
            oauth2_access_token=self._credentials['access_token'])
        self.localtime = localtime

    def __repr__(self):
        """Represent the dropbox filesystem and the root."""
        args = (self.__class__.__name__, self._root)
        return '<FileSystem: %s - Root Directory: %s>' % args

    __str__ = __repr__

    def __unicode__(self):
        """Represent the dropbox filesystem and the root (unicode)."""
        args = (self.__class__.__name__, self._root)
        return u'<FileSystem: %s - Root Directory: %s>' % args

    def getmeta(self, meta_name, default=NoDefaultMeta):
        """Get _meta info from DropboxFs."""
        if meta_name == 'read_only':
            return self.read_only
        return super(DropboxFS, self).getmeta(meta_name, default)

    def is_root(self, path):
        """Check if the given path is the root folder.

        :param path: Path to the folder to check
        """
        if(path == self._root):
            return True
        else:
            return False

    @synchronize
    def open(self, path, mode="rb", **kwargs):
        """Open the named file in the given mode.

        This method downloads the file contents into a local temporary
        file so that it can be worked on efficiently.  Any changes
        made to the file are only sent back to cloud storage when
        the file is flushed or closed.

        :param path: Path to the file to be opened
        :param mode: In which mode to open the file
        :raise ResourceNotFoundError: If given path doesn't exist and
            'w' is not in mode
        :return: RemoteFileBuffer object
        """
        path = abspath(normpath(path))
        spooled_file = SpooledTemporaryFile(mode=mode, bufsize=MAX_BUFFER)

        if "w" in mode:
            # Truncate the file if requested
            self.client.put_file(path, "", True)
        else:
            # Try to write to the spooled file, if path doesn't exist create it
            # if 'w' is in mode
            try:
                spooled_file.write(self.client.get_file(path).read())
                spooled_file.seek(0, 0)
            except:
                if "w" not in mode:
                    raise ResourceNotFoundError(path)
                else:
                    self.createfile(path, True)
        #  This will take care of closing the socket when it's done.
        return RemoteFileBuffer(self, path, mode, spooled_file)

    @synchronize
    def getcontents(self, path, mode="rb", **kwargs):
        """Get contents of a file."""
        path = abspath(normpath(path))
        return self.open(path, mode).read()

    def setcontents(self, path, data, *args, **kwargs):
        """Set new content to remote file.

        Method works only with existing files and sets
        new content to them.

        :param path: Path the file in which to write the new content
        :param contents: File contents as a string, or any object with
            read and seek methods
        :param kwargs: additional parameters like:
            encoding: the type of encoding to use if data is text
            errors: encoding errors
        :param chunk_size: Number of bytes to read in a chunk,
            if the implementation has to resort to a read copy loop
        :return: Path of the updated file

        """
        path = abspath(normpath(path))
        self.client.put_file(path, data, overwrite=True)
        return path

    def desc(self, path):
        """Get the title of a given path.

        :return: The title for the given path.
        """
        path = abspath(normpath(path))
        info = self.getinfo(path)
        return info["title"]

    def getsyspath(self, path, allow_none=False):
        """Return a path as the Dropbox API specifies."""
        if allow_none:
            return None
        return client.format_path(abspath(normpath(path)))

    def isdir(self, path):
        """Check if a the specified path is a folder.

        :param path: Path to the file/folder to check
        """
        info = self.getinfo(path)
        return info.get('isdir')

    def isfile(self, path):
        """Check if a the specified path is a file.

        :param path: Path to the file/folder to check
        """
        info = self.getinfo(path)
        return not info.get('isdir')

    def exists(self, path):
        """Check if a the specified path exists.

        :param path: Path to the file/folder to check
        """
        try:
            self.getinfo(path)
            return True
        except ResourceNotFoundError:
            return False

    def listdir(self, path="/", wildcard=None, full=False, absolute=False,
                dirs_only=False, files_only=False):
        """List the the files and directories under a given path.

        The directory contents are returned as a list of unicode paths

        :param path: path to the folder to list
        :type path: string
        :param wildcard: Only returns paths that match this wildcard
        :type wildcard: string containing a wildcard, or a callable
            that accepts a path and returns a boolean
        :param full: returns full paths (relative to the root)
        :type full: bool
        :param absolute: returns absolute paths
            (paths beginning with /)
        :type absolute: bool
        :param dirs_only: if True, only return directories
        :type dirs_only: bool
        :param files_only: if True, only return files
        :type files_only: bool
        :return: a list of unicode paths
        """
        path = abspath(normpath(path))
        children = self.client.children(path)
        return self._listdir_helper(path, children, wildcard, full, absolute,
                                    dirs_only, files_only)

    @synchronize
    def getinfo(self, path):
        """Get info from cloud service.

        Returned information is metadata from cloud service +
        a few more fields with standard names for some parts
        of the metadata.

        :param path: path to the file/folder for which to return
            informations
        :return: dictionary with informations about the specific file
        """
        path = abspath(normpath(path))
        metadata = self.client.metadata(path)
        return self._metadata_to_info(metadata, localtime=self.localtime)

    def copy(self, src, dst, *args, **kwargs):
        """Copy a file to another location.

        :param src: Path to the file to be copied
        :param dst: Path to the folder in which to copy the file
        :return: Path to the copied file
        """
        src = abspath(normpath(src))
        dst = abspath(normpath(dst))
        return self.client.file_copy(src, dst)

    def copydir(self, src, dst, *args, **kwargs):
        """Copy a directory to another location.

        :param src: Path to the folder to be copied
        :param dst: Path to the folder in which to copy the folder
        :return: Path to the copied folder
        """
        src = abspath(normpath(src))
        dst = abspath(normpath(dst))
        return self.client.file_copy(src, dst)

    def move(self, src, dst, chunk_size=16384, *args, **kwargs):
        """Move a file to another location.

        :param src: Path to the file to be moved
        :param dst: Path to the folder in which the file will be moved
        :param chunk_size: if using chunk upload
        :return: Path to the moved file
        """
        src = abspath(normpath(src))
        dst = abspath(normpath(dst))
        return self.client.file_move(src, dst)

    def movedir(self, src, dst, *args, **kwargs):
        """Move a directory to another location.

        :param src: Path to the folder to be moved
        :param dst: Path to the folder in which the folder will be moved
        :param chunk_size: if using chunk upload
        :return: Path to the moved folder
        """
        src = abspath(normpath(src))
        dst = abspath(normpath(dst))
        return self.client.file_move(src, dst)

    def rename(self, src, dst, *args, **kwargs):
        """Rename a file of a given path.

        :param src: Path to the file to be renamed
        :param dst: Full path with the new name
        :raise UnsupportedError: If trying to remove the root directory
        :return: Path to the renamed file
        """
        src = abspath(normpath(src))
        dst = abspath(normpath(dst))
        return self.client.file_move(src, dst)

    def makedir(self, path, recursive=False, allow_recreate=False):
        """Create a directory of a given path.

        :param path: path to the folder to be created.
            If only the new folder is specified
            it will be created in the root directory
        :param recursive: allows recursive creation of directories
        :param allow_recreate: dropbox currently doesn't support
            allow_recreate, so if a folder exists it will
        :return: Id of the created directory
        """
        if not self._checkRecursive(recursive, path):
            raise UnsupportedError("Recursively create specified folder.")

        path = abspath(normpath(path))
        return self.client.file_create_folder(path)

    def createfile(self, path, wipe=False, **kwargs):
        """Create an empty file.

        :param path: path to the new file.
        :param wipe: New file with empty content.
        :param kwargs: Additional parameters like description - a short
            description of the new file.

        :attention:

            Root directory is the current root directory of this instance of
            filesystem and not the root of your Google Drive.

        :return: Path to the created file
        """
        return self.client.put_file(path, '', overwrite=wipe)

    def remove(self, path):
        """Remove a file of a given path.

        :param path: path to the file to be deleted
        :return: None if removal was successful
        """
        path = abspath(normpath(path))
        if self.is_root(path=path):
            raise UnsupportedError("Can't remove the root directory")
        if self.isdir(path=path):
            raise ResourceInvalidError(
                "Specified path is a directory. Please use removedir.")

        self.client.file_delete(path)

    def removedir(self, path, *args, **kwargs):
        """Remove a directory of a given path.

        :param path: path to the file to be deleted
        :return: None if removal was successful
        """
        path = abspath(normpath(path))

        if self.is_root(path=path):
            raise UnsupportedError("Can't remove the root directory")
        if self.isfile(path=path):
            raise ResourceInvalidError(
                "Specified path is a directory. Please use removedir.")

        self.client.file_delete(path)

    def getpathurl(self, path):
        """Get the url of a given path.

        :param path: path to the file for which to return the url path
        :param allow_none: if true, this method can return None if
            there is no URL form of the given path
        :type allow_none: bool
        :return: url that corresponds to the given path, if one exists

        """
        path = abspath(normpath(path))
        return self.client.media(path)

    def about(self):
        """Get info about the current user.

        :return: information about the current user
            with whose credentials is the file system instantiated.
        """
        info = self.client.account_info()
        info['cloud_storage_url'] = "http://www.dropbox.com/"
        info['user_name'] = info.pop('display_name')
        info['quota'] = 100 * (info['quota_info']["normal"] + info['quota_info']
                               ["shared"]) / float(info['quota_info']["quota"])
        return info
        return self.client.account_info()

    def _checkRecursive(self, recursive, path):
        #  Checks if the new folder to be created is compatible with current
        #  value of recursive
        parts = path.split("/")
        if(parts < 3):
            return True

        testPath = "/".join(parts[:-1])
        if self.exists(testPath):
            return True
        elif recursive:
            return True
        else:
            return False

    def _metadata_to_info(self, metadata, localtime=False):
        """Return modified metadata.

        Method adds a few standard names to the metadata:
            size - the size of the file/folder
            isdir - is something a file or a directory
            created_time - the time of the creation
            path - path to the object which metadata are we showing
            revision - google drive doesn't have a revision parameter
            modified - time of the last modification
        :return: The full metadata and a few more fields
            with standard names.
        """
        info = {
            'size': metadata.get('bytes', 0),
            'isdir': metadata.get('is_dir', False),
            'title': metadata['path'].split("/")[-1],
            'created_time': None
        }
        try:
            mtime = metadata.get('modified', None)
            if mtime:
                # Parse date/time from Dropbox as struct_time.
                mtime = time.strptime(mtime, TIME_FORMAT)
                if localtime:
                    # Convert time to local timezone in seconds.
                    mtime = calendar.timegm(mtime)
                else:
                    mtime = time.mktime(mtime)
                # Convert to datetime object, store in modified_time
                info['modified'] = datetime.datetime.fromtimestamp(mtime)
        except KeyError:
            pass

        info.update(metadata)
        return info



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Babel datetime localization template filters for Jinja.

See full documentation of corresponding methods in Flask-Babel:
https://pythonhosted.org/Flask-Babel/
"""

from flask_babel import format_date, format_datetime, format_time, \
    format_timedelta, to_user_timezone, to_utc


def filter_to_user_timezone(dt):
    """Convert a datetime object to the user's timezone."""
    return to_user_timezone(dt)


def filter_to_utc(dt):
    """Convert a datetime object to UTC and drop tzinfo."""
    return to_utc(dt)


def filter_format_datetime(dt, format=None, rebase=True):
    """Return a date formatted according to the given pattern.

    The format parameter can either be ``short``, ``medium``, ``long`` or
    ``full``.
    """
    return format_datetime(dt, format=format, rebase=rebase)


def filter_format_date(dt, format=None, rebase=True):
    """Return a date formatted according to the given pattern."""
    return format_date(dt, format=format, rebase=rebase)


def filter_format_time(dt, format=None, rebase=True):
    """Return a date formatted according to the given pattern."""
    return format_time(dt, format=format, rebase=rebase)


def filter_format_timedelta(dt, format=None, rebase=True):
    """Return a date formatted according to the given pattern."""
    return format_timedelta(dt, format=format, rebase=rebase)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Administration menu."""

from flask_login import current_user
from flask_menu import Menu

from invenio_base.i18n import _

menu = Menu()


def setup_app(app):
    """Register all subitems to the 'main.admin' menu item."""
    menu.init_app(app)

    @app.before_first_request
    def register_item():
        item = app.extensions['menu'].submenu('main.admin')
        item.register('admin.index', _('Admin'), order=10,
                      visible_when=lambda: current_user.is_admin)



# -*- coding: utf-8 -*-
# This file is part of Invenio.
# Copyright (C) 2012, 2013, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
"""
    invenio_ext.legacy.request_class
    --------------------------------

     This module provides Flask Request wrapper.
"""

from flask import Request


class LegacyRequest(Request):
    """
    Flask Request wrapper which adds support for converting the request into
    legacy request (SimulatedModPythonRequest). This is primarily useful
    when Flaskifying modules, that still depends on old code.
    """

    def dummy_start_response(self, *args, **kwargs):
        """Dummy function for simulating response begining."""
        pass

    def get_legacy_request(self):
        """Returns an instance of SimulatedModPythonRequest."""
        from invenio.legacy.wsgi import SimulatedModPythonRequest
        return SimulatedModPythonRequest(self.environ,
                                         self.dummy_start_response)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""CrossRef extension.

CrossRef extension is initialized like this:

>>> from flask import Flask
>>> from invenio_ext.crossref import CrossRef
>>> app = Flask("myapp")
>>> ext = CrossRef(app=app)

Configuration Settings
----------------------
The details of the CrossRef URL and endpoint can be customized in the application
settings.

========================= =====================================================
`CROSSREF_API_URL`        The URL of CrossRef query API.
                          **Default:** `http://api.crossref.org/works`
`CROSSREF_ENDPOINT`       The name of Flask endpoint for new application route.
                          If the value is `False` (or `None`) the url rule is
                          not registered.  **Default:** `_crossref.search`
`CROSSREF_URL_RULE`       The URL for `CROSSREF_ENPOINT` (i.e.
                          `url_for(current_app.config['CROSSREF_ENDPOINT'])`
                          is equal TO `current_app.config['CROSSREF_ENDPOINT']`
                          ). **Default:** `/crossref/search`
`CROSSREF_SEARCH_PREFIX`  The prefix used to perform the database search.
========================= =====================================================

"""

from __future__ import absolute_import

import requests
from urlparse import urljoin

from flask import current_app, request, jsonify
from flask_login import login_required

response_code = {'success': 200,
                 'notfound': 404,
                 'malformed': 422,
                 'multiplefound': 300}


class CrossRef(object):

    """CrossRef extension implementation.

    Initialization of the extension:

    >>> from flask import Flask
    >>> from invenio_ext.crossref import CrossRef
    >>> app = Flask("myapp")
    >>> ext = CrossRef(app=app)

    or alternatively using the factory pattern:

    >>> app = Flask("myapp")
    >>> ext = CrossRef()
    >>> ext.init_app(app)
    """

    def __init__(self, app=None):
        """Initialize extension."""
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """Initialize a Flask application."""
        app.config.setdefault("CROSSREF_API_URL",
                              "http://api.crossref.org/works/")
        app.config.setdefault("CROSSREF_ENDPOINT", "_doi.search")
        app.config.setdefault("CROSSREF_URL_RULE", "/doi/search")

        # Follow the Flask guidelines on usage of app.extensions
        if not hasattr(app, "extensions"):
            app.extensions = {}
        if "crossref" in app.extensions:
            raise RuntimeError("Flask application already initialized")
        app.extensions["crossref"] = self

        if app.config["CROSSREF_ENDPOINT"]:
            app.add_url_rule(app.config["CROSSREF_URL_RULE"],
                             app.config["CROSSREF_ENDPOINT"],
                             login_required(self.search))

    def get_response(self, crossref_doi):
        """Get CrossRef response from the ``CROSSREF_API_URL`` page."""
        response = requests.get(
            urljoin(
                current_app.config["CROSSREF_API_URL"],
                "{term}".format(term=crossref_doi.strip()),
            ),
        )
        return response

    def get_json(self, doi):
        """Get CrossRef json data."""
        response = self.get_response(doi)
        data, query = {}, {}

        if response.status_code == 200:
            if 'message' in response.json():
                query = response.json().get('message')
            else:
                query = response.json()
            data["status"] = "success"
        elif response.status_code == 404:
            data["status"] = "notfound"

        data["source"] = "crossref"
        data["query"] = query

        return data

    def search(self, doi=None):
        """Search for given DOI."""
        doi = (self.app.config.get("CROSSREF_SEARCH_PREFIX", "") + doi if doi
               else request.args.get("doi"))

        from invenio_records.utils import get_unique_record_json

        # query the database
        result = get_unique_record_json(doi)
        if result["status"] == "notfound":
            # query crossref
            result = self.get_json(doi)

        resp = jsonify(result)
        resp.status_code = response_code.get(result['status'], 200)
        return resp

__all__ = ("CrossRef", )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2011, 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 52 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Platform-independent TinyInteger type."""

from sqlalchemy.dialects.mysql import TINYINT
from sqlalchemy.types import Integer, TypeDecorator


class LegacyTinyInteger(TypeDecorator):

    """Platform-independent TinyInteger type.

    Uses MySQL's :class:`~sqlalchemy.dialects.mysql.TINYINT` type, otherwise
    uses SQLAlchemy definition of :class:`~sqlalchemy.types.Integer`.
    """

    impl = Integer

    def __init__(self, display_width=2, unsigned=False, **kwargs):
        """Reserve special arguments only for MySQL Platform."""
        self.display_width = display_width
        self.unsigned = unsigned
        super(LegacyTinyInteger, self).__init__(**kwargs)

    def load_dialect_impl(self, dialect):
        """Load dialect dependent implementation."""
        if dialect.name == 'mysql':
            return dialect.type_descriptor(TINYINT(self.display_width,
                                                   unsigned=self.unsigned))
        else:
            return dialect.type_descriptor(Integer)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2013, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""
    invenio_ext.sslify.decorators
    -----------------------------

    Implements custom decorators.
"""


def ssl_required(f):
    """This forces the view function be available only through HTTPS."""
    f._force_https = True
    return f



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""
Debug Toolbar Extension.

Configuration
-------------
The toolbar supports several configuration options:

====================================  =====================================   ==========================
Name                                  Description                             Default
====================================  =====================================   ==========================
``DEBUG_TB_ENABLED``                  Enable the toolbar?                     ``app.debug``
``DEBUG_TB_HOSTS``                    Whitelist of hosts to display toolbar   any host
``DEBUG_TB_INTERCEPT_REDIRECTS``      Should intercept redirects?             ``True``
``DEBUG_TB_PANELS``                   List of module/class names of panels    enable all built-in panels
``DEBUG_TB_PROFILER_ENABLED``         Enable the profiler on all requests     ``False``, user-enabled
``DEBUG_TB_TEMPLATE_EDITOR_ENABLED``  Enable the template editor              ``False``
====================================  =====================================   ==========================

For more information see http://flask-debugtoolbar.readthedocs.org/en/latest/
"""


def setup_app(app):
    """Setup Flask with the DebugToolbar application."""
    # Enable Flask Debug Toolbar early to also catch HTTPS redirects
    if app.debug and app.config.get('DEBUG_TB_ENABLED', app.debug):
        try:
            from flask_debugtoolbar import module, DebugToolbarExtension
            module.static_folder = 'static'
            DebugToolbarExtension(app)
        except ImportError:
            app.logger.exception("Flask-DebugToolbar is missing")

    return app



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Google Drive file system.

Installation:: pip install invenio[google_drive]

Known issues:

* Flush and close, both call write contents and because of that
  the file on cloud is overwrite twice...
"""

import datetime
import os
import time
from UserDict import UserDict

import httplib2
import six
from apiclient import errors
from apiclient.discovery import build
from apiclient.http import MediaInMemoryUpload
# Python filesystem imports
from fs.base import FS
from fs.errors import CreateFailedError, NoPathURLError, \
    OperationFailedError, RemoteConnectionError, ResourceInvalidError, \
    ResourceNotFoundError, UnsupportedError
from fs.filelike import SpooledTemporaryFile
from fs.remote import RemoteFileBuffer
from oauth2client.client import OAuth2Credentials

# Items in cache are considered expired after 5 minutes.
CACHE_TTL = 300
# Max size for spooling to memory before using disk (5M).
MAX_BUFFER = 1024**2 * 5
# Indicates the mimeType for a google drive folder
GD_FOLDER = "application/vnd.google-apps.folder"


class CacheItem(object):

    """Represents a path in the cache.

    There are two components to a path. It's individual metadata,
    and the children contained within it.
    """

    def __init__(self, metadata=None, children=None, timestamp=None,
                 parents=None):
        """Initialize a CacheItem instance."""
        self.metadata = metadata
        self.children = children
        self.parents = parents
        if timestamp is None:
            timestamp = time.time()
        self.timestamp = timestamp

    def add_child(self, name, client=None):
        """Add a child."""
        if self.children is None:
            if client is not None:
                # This is a fix. When you add a child to a folder that
                # was still not listed, that folder gets only one
                # child when you list it afterwards. So this fix
                # first tries to check are the files/folders inside
                # this directory on cloud.
                client.children(self.metadata['id'])
            else:
                self.children = [name]
        else:
            if name not in self.children:
                self.children.append(name)

    def del_child(self, name):
        """Delete a child."""
        if self.children is None:
            return
        try:
            i = self.children.index(name)
        except ValueError:
            return
        self.children.pop(i)

    def _get_expired(self):
        if self.timestamp <= time.time() - CACHE_TTL:
            return True
    expired = property(_get_expired)

    def renew(self):
        """Renew."""
        self.timestamp = time.time()


class GoogleDriveCache(UserDict):

    """Represent the google drive cache."""

    def __init__(self, client):
        """Initialize a GoogleDriveCache instance."""
        self._client = client
        UserDict.__init__(self)

    def set(self, path, metadata, children=None, parents=None):
        """Set metadata."""
        self[path] = CacheItem(metadata, children=children, parents=parents)
        if parents is not None:
            for parent in parents:
                if parent in self:
                    self.get(parent).add_child(path, self._client)

    def pop(self, path, default=None):
        """Pop data of a given path."""
        value = UserDict.pop(self, path, default)
        if value.parents is not None:
            for parent in value.parents:
                if parent in self:
                    self.get(parent).del_child(value.metadata['id'])
        return value


class GoogleDriveClient(object):

    """Represent the google drive client."""

    def __init__(self, credentials):
        """Initialize a GoogleDriveClient instance."""
        self.credentials = credentials
        self.service = self._build_service()
        self.cache = GoogleDriveCache(self)
        self._retry = 0

    def _build_service(self):
        http = httplib2.Http()
        http = self.credentials.authorize(http)
        service = build('drive', 'v2', http=http)
        return service

    def _retry_operation(self, method, *args):
        """Method retries an operation.

        Sometimes access_token expires and we need to rebuild it using
        the refresh token. This method does that and retries the
        operation that failed.
        """
        if self._retry < 5:
            self._retry += 1
            self.service = self._build_service()
            return method(*args)
        else:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired "
                "or user credentials are invalid.")

    def get_file(self, path):
        """Get file of a given path."""
        item = self.cache.get(path)
        if not item or item.metadata is None or item.expired:
            try:
                metadata = self.service.files().get(fileId=path).execute()
            except errors.HttpError as e:
                if e.resp.status == 404:
                    raise ResourceNotFoundError("Source file doesn't exist")
                raise OperationFailedError(opname='get_file',
                                           msg=e.resp.reason)
            except:
                return self._retry_operation(self.get_file, path)
            self._add_to_cache_from_dict(metadata, metadata['parents'])
        else:
            item = self.cache[path]
            metadata = item.metadata

        download_url = metadata.get('downloadUrl')
        resp, content = self.service._http.request(download_url)
        if resp.status == 200:
            return content
        else:
            raise OperationFailedError(opname="get_file", msg=str(resp))

    def metadata(self, path):
        """Get metadata of a given path."""
        item = self.cache.get(path)
        if not item or item.metadata is None or item.expired:
            try:
                metadata = self.service.files().get(fileId=path).execute()
            except errors.HttpError as e:
                if e.resp.status == 404:
                    raise ResourceNotFoundError(path)
                raise OperationFailedError(opname='metadata', path=path,
                                           msg=e.resp.reason)
            except:
                return self._retry_operation(self.metadata, path)
            if metadata.get('trashed', False):
                raise ResourceNotFoundError(path)
            item = self.cache[path] = CacheItem(metadata)

        # Copy the info so the caller cannot affect our cache.
        return dict(item.metadata.items())

    def children(self, path):
        """Get children of a given path."""
        update = False
        item = self.cache.get(path)
        if item:
            if item.expired:
                update = True
            else:
                if item.metadata["mimeType"] != GD_FOLDER:
                    raise ResourceInvalidError(path)
            if not item.children:
                update = True
        else:
            update = True
        if update:
            try:
                metadata = self.service.files().get(fileId=path).execute()
                if metadata["mimeType"] != GD_FOLDER:
                    raise ResourceInvalidError(path)
                param = {"q": "trashed=false and '%s' in parents" % path}
                children = []
                filesResource = self.service.files().list(**param).execute()
                for child in filesResource['items']:
                    children.append(child['id'])
                    self.cache[child['id']] = CacheItem(child, parents=[path])
                item = self.cache[path] = CacheItem(metadata, children)
            except errors.HttpError as e:
                if e.resp.status == 404:
                    raise ResourceNotFoundError(path)
                if not item or e.resp.status != 304:
                    raise OperationFailedError(opname='metadata', path=path,
                                               msg=e.resp.reason)
                # We have an item from cache (perhaps expired),
                # but it's still valid (as far as GoogleDrive is
                # concerned), so just renew it and keep using it.
                item.renew()
            except:
                return self._retry_operation(self.children, path)
        return item.children

    def file_create_folder(self, parent_id, title):
        """Add newly created directory to cache."""
        body = {
            "title": title,
            "parents": [{"id": parent_id}],
            "mimeType": "application/vnd.google-apps.folder"
        }
        try:
            metadata = self.service.files().insert(body=body).execute()
        except errors.HttpError as e:
            if e.resp.status == 405:
                raise ResourceInvalidError(parent_id)
            if e.resp.status == 404:
                raise ResourceNotFoundError(parent_id)
            raise OperationFailedError(
                opname='file_create_folder',
                msg="%s, the reasons could be: parent "
                    "doesn't exist or is a file" % (e.resp.reason, ))
        except:
            return self._retry_operation(self.file_create_folder,
                                         parent_id, title)

        self.cache.set(metadata["id"], metadata, parents=[parent_id])
        return metadata

    def file_copy(self, file_id, parent_id):
        """Copy a file to a folder of a given id."""
        body = {"parents": [{"id": parent_id}]}

        try:
            metadata = self.service.files().copy(fileId=file_id,
                                                 body=body,
                                                 ).execute()
        except errors.HttpError as e:
            if e.resp.status == 404:
                raise ResourceNotFoundError(
                    "Parent or source file don't exist.")
            raise OperationFailedError(opname='file_copy', msg=e.resp.reason)
        except:
            return self._retry_operation(self.file_copy, file_id, parent_id)

        self.cache.set(metadata['id'], metadata, parents=[parent_id])
        return metadata

    def update_file(self, file_id, new_file):
        """Update the file's contents of the given id."""
        try:
            metadata = self.service.files().update(fileId=file_id,
                                                   body=new_file
                                                   ).execute()
        except errors.HttpError as e:
            if e.resp.status == 404:
                raise ResourceNotFoundError(
                    "Parent or source file don't exist.")
            raise OperationFailedError(opname='update_file',
                                       msg=e.resp.reason)
        except:
            return self._retry_operation(self.update_file, file_id, new_file)
        self.cache.pop(file_id, None)
        self._add_to_cache_from_dict(metadata, metadata['parents'])
        return metadata

    def _add_to_cache_from_dict(self, metadata, parents):
        new_parents = []
        for one in parents:
            new_parents.append(one['id'])
        self.cache.set(metadata['id'], metadata, parents=new_parents)

    def update_file_content(self, file_id, content):
        """Update the file's contents of the given id on google drive."""
        item = self.cache.get(file_id, None)
        if not item or item.metadata is None or item.expired:
            try:
                metadata = self.service.files().get(fileId=file_id).execute()
            except errors.HttpError as e:
                raise OperationFailedError(opname='update_file_content',
                                           msg=e.resp.reason)
            except:
                return self._retry_operation(self.update_file_content,
                                             file_id, content)
            self.cache.set(metadata['id'], metadata)
        else:
            metadata = item.metadata

        media_body = MediaInMemoryUpload(content)
        try:
            updated_file = self.service.files().update(fileId=file_id,
                                                       body=metadata,
                                                       media_body=media_body
                                                       ).execute()
        except errors.HttpError as e:
            raise OperationFailedError(opname='update_file_content',
                                       msg=e.resp.reason)
        except TypeError as e:
            raise ResourceInvalidError("update_file_content %r" % e)
        except:
            return self._retry_operation(self.update_file_content, file_id,
                                         content)
        self.cache.pop(file_id, None)
        self._add_to_cache_from_dict(updated_file, updated_file['parents'])
        return updated_file

    def file_delete(self, path):
        """Delete a file."""
        try:
            self.service.files().delete(fileId=path).execute()
        except errors.HttpError as e:
            if e.resp.status == 404:
                raise ResourceNotFoundError(path)
            raise OperationFailedError(opname='file_delete',
                                       msg=e.resp.reason)
        except:
            return self._retry_operation(self.file_delete, path)
        self.cache.pop(path, None)

    def put_file(self, parent_id, title, content, description=None):
        """Add a file to folder of the given id."""
        media_body = MediaInMemoryUpload(content)
        body = {
            'title': title,
            'description': description,
            'parents': [{'id': parent_id}]
        }
        try:
            metadata = self.service.files().insert(body=body,
                                                   media_body=media_body
                                                   ).execute()
        except errors.HttpError as e:
            raise OperationFailedError(opname='put_file', msg=e.resp.reason)
        except TypeError as e:
            raise ResourceInvalidError("put_file")
        except:
            return self._retry_operation(self.put_file, parent_id, title,
                                         content, description)
        self.cache.set(metadata['id'], metadata, parents=[parent_id])
        return metadata

    def about(self):
        """Get info about the service."""
        try:
            # FIXME check if the access_token has expired
            info = self.service.about().get().execute()
            return info
        except:
            return self._retry_operation(self.about)


class GoogleDriveFS(FS):

    """Google drive file system.

    @attention: when setting variables in os.environ please note that
    GD_TOKEN_EXPIRY has to be in format: "%Y, %m, %d, %H, %M, %S, %f"
    """

    __name__ = "Google Drive"
    _meta = {'thread_safe': True,
             'virtual': False,
             'read_only': False,
             'unicode_paths': True,
             'case_insensitive_paths': False,
             'network': True,
             'atomic.move': True,
             'atomic.copy': True,
             'atomic.makedir': True,
             'atomic.rename': True,
             'atomic.setcontents': True,
             }

    def __init__(self, root=None, credentials=None, thread_synchronize=True):
        """Initialize a GoogleDriveFS instance."""
        self._root = root

        def _getDateTimeFromString(time):
            # Parses string into datetime object
            if time:
                return datetime.datetime.strptime(
                    time, "%Y, %m, %d, %H, %M, %S, %f")
            else:
                return None

        if credentials is None:
            # Get credentials need to build the google drive service
            if ("GD_ACCESS_TOKEN" not in os.environ or
                    "GD_CLIENT_ID" not in os.environ or
                    "GD_CLIENT_SECRET" not in os.environ or
                    "GD_TOKEN_EXPIRY" not in os.environ or
                    "GD_TOKEN_URI" not in os.environ,
                    "GD_REFRESH_TOKEN" not in os.environ):
                raise CreateFailedError("You need to set:\n"
                                        "GD_ACCESS_TOKEN, "
                                        "GD_CLIENT_ID, "
                                        "GD_CLIENT_SECRET "
                                        "GD_TOKEN_EXPIRY, "
                                        "GD_REFRESH_TOKEN, "
                                        "GD_TOKEN_URI in os.environ"
                                        )
            else:
                self._credentials = OAuth2Credentials(
                    os.environ.get('GD_ACCESS_TOKEN'),
                    os.environ.get('GD_CLIENT_ID'),
                    os.environ.get('GD_CLIENT_SECRET'),
                    os.environ.get('GD_REFRESH_TOKEN'),
                    _getDateTimeFromString(
                        os.environ.get('GD_TOKEN_EXPIRY')
                    ),
                    os.environ.get('GD_TOKEN_URI'),
                    None
                )
        else:
            self._credentials = OAuth2Credentials(
                credentials.get('access_token'),
                credentials.get('client_id'),
                credentials.get('client_secret'),
                credentials.get('refresh_token'),
                _getDateTimeFromString(
                    credentials.get('token_expiry')
                ),
                credentials.get('token_uri'),
                None
            )
        self.client = GoogleDriveClient(self._credentials)

        if self._root is None or root == '' or self._root == "/":
            # Root fix, if root is not set get the root folder id
            # from the user information returned by 'about'
            about = self.client.about()
            self._root = about.get("rootFolderId")

        # Initialize super class FS
        super(self.__class__, self).__init__(
            thread_synchronize=thread_synchronize
        )

    def __repr__(self):
        """Represent the google_drive filesystem and the root."""
        args = (self.__class__.__name__, self._root)
        return '<FileSystem: %s - Root Directory: %s>' % args

    __str__ = __repr__

    def __unicode__(self):
        """Represent the google_drive filesystem and the root (unicode)."""
        args = (self.__class__.__name__, self._root)
        return u'<FileSystem: %s - Root Directory: %s>' % args

    def _update(self, path, contents):
        """Update contents of an existing file.

        :param path: Id of the file for which to update content
        :param contents: Contents to write to the file
        :return: Id of the updated file
        """
        path = self._normpath(path)

        if isinstance(contents, six.string_types):
            string_data = contents
        else:
            try:
                contents.seek(0)
                string_data = contents.read()
            except:
                raise ResourceInvalidError("Unsupported type")

        return self.client.update_file_content(path, string_data)['id']

    def setcontents(self, path, contents="", chunk_size=64 * 1024, **kwargs):
        """Set new content to remote file.

        Method works only with existing files and sets
        new content to them.

        :param path: Id of the file in which to write the new content
        :param contents: File contents as a string, or any object with
            read and seek methods
        :param kwargs: additional parameters like:
            encoding: the type of encoding to use if data is text
            errors: encoding errors
        :param chunk_size: Number of bytes to read in a chunk,
            if the implementation has to resort to a read copy loop
        :return: Id of the updated file
        """
        encoding = kwargs.get("encoding", None)
        errors = kwargs.get("errors", None)

        if isinstance(contents, six.text_type):
            contents = contents.encode(encoding=encoding, errors=errors)

        return self._update(path, contents)

    def createfile(self, path, wipe=True, **kwargs):
        """Create always an empty file.

        Even if another file with the same name exists it will create
        a file with the same name.

        :param path: path to the new file. It has to be in one of
            following forms: parent_id/file_title.ext, file_title.ext or
            /file_title.ext - in this cases root directory is the parent.
        :param wipe: New file with empty content.
            In the case of google drive it will always be True
        :param kwargs: Additional parameters like:
            description - a short description of the new file
        :raises ResourceNotFoundError: If parent doesn't exist.

        :attention: Root directory is the current root directory
            of this instance of filesystem and not the root of
            your Google Drive.

        :return: Id of the created file
        """
        # Google drive doesn't work with paths. So a slight
        # work around is needed.
        path = self._normpath(path)

        parts = path.split("/")
        description = kwargs.get('description', '')
        if len(parts) == 1 and parts[0] == self._root:
            raise ResourceNotFoundError("Please, specify a filename.")
        elif len(parts) == 1:
            return self.client.put_file(self._root,
                                        parts[0], "", description)['id']
        else:
            title = parts.pop()
            parent_id = self.makedir(parts, True)
            return self.client.put_file(parent_id,
                                        title, "", description)['id']

    def open(self, path, mode='r', buffering=-1, encoding=None,
             errors=None, newline=None, line_buffering=False, **kwargs):
        """Open the named file in the given mode.

        This method downloads the file contents into a local temporary file so
        that it can be worked on efficiently.  Any changes made to the file are
        only sent back to cloud storage when the file is flushed or closed.

        :param path: Id of the file to be opened
        :param mode: In which mode to open the file
        :raises ResourceNotFoundError: If given path doesn't exist and
            'w' is not in mode
        :return: RemoteFileBuffer object
        """
        path = self._normpath(path)
        spooled_file = SpooledTemporaryFile(mode=mode, bufsize=MAX_BUFFER)

        # Truncate the file if requested
        if "w" in mode:
            try:
                self._update(path, "")
            except:
                path = self.createfile(path, True)
        else:
            try:
                spooled_file.write(self.client.get_file(path))
                spooled_file.seek(0, 0)
            except Exception as e:
                if "w" not in mode and "a" not in mode:
                    raise ResourceNotFoundError("%r" % e)
                else:
                    path = self.createfile(path, True)
        return RemoteFileBuffer(self, path, mode, spooled_file)

    def is_root(self, path):
        """Check if the given path is the root folder.

        :param path: Id of the folder to check
        """
        path = self._normpath(path)

        if path == self._root:
            return True
        else:
            return False

    def copy(self, src, dst, overwrite=False, chunk_size=1024 * 64):
        """Copy a file to another folder.

        :param src: Id of the file to be copied
        :param dst: Id of the folder in which to copy the file
        :param overwrite: This is never true for GoogleDrive
            because there can be many files with the same name
            in one folder.
        :return: Id of the copied file
        """
        return self.client.file_copy(src, dst)['id']

    def copydir(self, src, dst, overwrite=False, ignore_errors=False,
                chunk_size=16384):
        """Copy of folders is not supported.

        @attention: Google drive doesn't support copy of folders.
        To implement it over copy method will be very inefficient.
        """
        raise NotImplemented("If implemented method will be very inefficient")

    def rename(self, src, dst):
        """Rename a file of a given path.

        :param src: Id of the file to be renamed
        :param dst: New title of the file
        :raises UnsupportedError: If trying to rename the root directory
        :return: Id of the renamed file
        """
        if self.is_root(path=src):
            raise UnsupportedError("Can't rename the root directory")
        f = self.client.metadata(src)
        f['title'] = dst
        return self.client.update_file(src, f)['id']

    def remove(self, path):
        """Remove a file of a given path.

        :param path: id of the file to be deleted
        :return: None if removal was successful
        """
        path = self._normpath(path)

        if self.is_root(path=path):
            raise UnsupportedError("Can't remove the root directory")
        if self.isdir(path=path):
            raise ResourceInvalidError("Specified path is a directory. "
                                       "Please use removedir.")
        self.client.file_delete(path)

    def removedir(self, path):
        """Remove a directory of a given path.

        :param path: id of the folder to be deleted
        :return: None if removal was successful
        """
        path = self._normpath(path)

        if not self.isdir(path):
            raise ResourceInvalidError("Specified path is not a directory")
        if self.is_root(path=path):
            raise UnsupportedError("remove the root directory")
        self.client.file_delete(path)

    def makedir(self, path, recursive=False, allow_recreate=False):
        """Create a directory of a given path.

        :param path: path to the folder you want to create.
            It has to be in one of the following forms:
            ``parent_id/new_folder_name`` when recursive is False,
            ``parent_id/new_folder1/new_folder2`` when recursive is True,
            ``/new_folder_name`` to create a new folder in root directory,
            ``/new_folder1/new_folder2`` to recursively create a new folder
            in root directory.
        :param recursive: allows recursive creation of directories
        :param allow_recreate: for google drive this param is
            always False, it will never recreate a directory with
            the same id ( same names are allowed )
        :return: Id of the created directory
        """
        path = self._normpath(path)
        if '/' in path:
            parts = path.split("/")
        else:
            parts = path

        if len(parts) == 1 and parts[0] == self._root:
            raise ResourceNotFoundError("Please, specify a folder name.")
        elif len(parts) == 1:
            return self.client.file_create_folder(self._root, parts[0])['id']
        else:
            if recursive:
                resp = self.client.file_create_folder(self._root, parts[0])
                for folder in parts[1:]:
                    if folder != '':
                        resp = self.client.file_create_folder(resp['id'],
                                                              folder)
                return resp['id']
            else:
                raise UnsupportedError("Recursively create a folder.")

    def move(self, src, dst, overwrite=False, chunk_size=16384):
        """Move a file to another folder.

        .. note:: Google Drive can have many parents for one file,
            when using this method a file will be moved from all
            current parents to the new parent 'dst'.

        :param src: id of the file to be moved
        :param dst: id of the folder in which the file will be moved
        :param overwrite: for Google drive it is always false
        :param chunk_size: if using chunk upload
        :return: Id of the moved file
        """
        if self.isdir(src):
            raise ResourceInvalidError(
                "Specified src is a directory. Please use movedir.")

        f = self.client.get_file(src)
        f['parents'] = [{"id": dst}]
        return self.client.update_file(src, f)['id']

    def movedir(self, src, dst, overwrite=False, ignore_errors=False,
                chunk_size=16384):
        """Move a folder to another folder.

        .. note:: Google Drive can have many parents for one folder,
            when using this method a folder will be moved from all
            current parents to the new parent 'dst'.

        :param src: id of the folder to be moved
        :param dst: id of the folder in which the file will be moved
        :param overwrite: for Google drive it is always false
        :param chunk_size: if using chunk upload
        :return: Id of the moved folder
        """
        if self.isfile(src):
            raise ResourceInvalidError(
                "Specified src is a file. Please use move.")
        f = self.client.get_file(src)
        f['parents'] = [{"id": dst}]

        return self.client.update_file(src, f)['id']

    def isdir(self, path):
        """Check if a the specified path is a directory.

        :param path: Id of the file/folder to check
        """
        path = self._normpath(path)
        info = self.getinfo(path)
        return info['isdir']

    def isfile(self, path):
        """
        Check if a the specified path is a file.

        :param path: Id of the file/folder to check
        """
        path = self._normpath(path)
        info = self.getinfo(path)
        return not info['isdir']

    def exists(self, path):
        """Check if a the specified path exists.

        :param path: Id of the file/folder to check
        """
        path = self._normpath(path)
        try:
            self.client.metadata(path)
            return True
        except RemoteConnectionError as e:
            raise e
        except:
            return False

    def listdir(self, path=None, wildcard=None, full=False, absolute=False,
                dirs_only=False, files_only=False, overrideCache=False):
        """List the files and directories under a given path.

        The directory contents are returned as a list of unicode paths

        :param path: id of the folder to list
        :type path: string
        :param wildcard: Only returns paths that match this wildcard
        :type wildcard: string containing a wildcard, or a callable
            that accepts a path and returns a boolean
        :param full: returns full paths (relative to the root)
        :type full: bool
        :param absolute: returns absolute paths
            (paths beginning with /)
        :type absolute: bool
        :param dirs_only: if True, only return directories
        :type dirs_only: bool
        :param files_only: if True, only return files
        :type files_only: bool
        :return: a list of unicode paths
        """
        path = self._normpath(path)
        flist = self.client.children(path)

        dirContent = self._listdir_helper('', flist, wildcard, full,
                                          absolute, dirs_only, files_only)
        return dirContent

    def listdirinfo(self, path=None, wildcard=None, full=False, absolute=False,
                    dirs_only=False, files_only=False):
        """Retrieve a list of paths and path info under a given path.

        This method behaves like listdir() but instead of just returning the
        name of each item in the directory, it returns a tuple of the name and
        the info dict as returned by getinfo.

        :param path: id of the folder
        :param wildcard: filter paths that match this wildcard
        :param dirs_only: only retrieve directories
        :type dirs_only: bool
        :param files_only: only retrieve files
        :type files_only: bool
        :return: tuple of the name and the info dict as
            returned by getinfo.
        """
        return [(p, self.getinfo(p)) for p in self.listdir(
            path, wildcard=wildcard, full=full, absolute=absolute,
            dirs_only=dirs_only, files_only=files_only)]

    def getinfo(self, path):
        """Returned information is metadata from cloud service.

        A few more fields with standard names for some parts of the metadata.
        :param path: file id for which to return informations
        :return: dictionary with informations about the specific file
        """
        path = self._normpath(path)
        return self._metadata_to_info(self.client.metadata(path))

    def getpathurl(self, path, allow_none=False):
        """Get the url of a given path.

        :param path: id of the file for which to return the url path
        :param allow_none: if true, this method can return None if
            there is no URL form of the given path
        :type allow_none: bool
        :raises `fs.errors.NoPathURLError`: If no URL form exists,
            and allow_none is False (the default)
        :return: url that corresponds to the given path, if one exists
        """
        url = None
        try:
            url = self.getinfo(path)
            url = url["webContentLink"]
        except RemoteConnectionError as e:
            raise e
        except:
            if not allow_none:
                raise NoPathURLError(path=path)

        return url

    def desc(self, path):
        """Get the title of a given path.

        :return: The title for the given path.
        """
        info = self.getinfo(path)
        return info["title"]

    def about(self):
        """Get user information and settings.

        :return: information about the current user
            with whose credentials is the file system instantiated.
        """
        info = self.client.about()
        info['cloud_storage_url'] = "http://drive.google.com/"
        info['user_name'] = info.get('name')
        info['quota'] = 100 * (float(info['quotaBytesUsed']) /
                               float(info['quotaBytesTotal']))
        return info

    def _normpath(self, path):
        """Method normalises the path for google drive.

        :return: normaliesed path as a string
        """
        if path is None or path == "" or path == '/' or len(path) == 0:
            return self._root
        elif path[0] == "/":
            return path[1:]
        else:
            return path

    def _metadata_to_info(self, metadata, localtime=False):
        """Return modified metadata.

        Method adds a few standard names to the metadata:
        * size - the size of the file/folder
        * isdir - is something a file or a directory
        * created_time - the time of the creation
        * path - path to the object which metadata are we showing
        * revision - google drive doesn't have a revision parameter
        * modified - time of the last modification

        :return: The full metadata and a few more fields
            with standard names.
        """
        isdir = metadata.get("mimeType", None) == GD_FOLDER
        info = {
            'size': metadata.get('fileSize', 0),
            'isdir': isdir,
            'created_time': metadata.get('createdDate', 0),
            'path': metadata.get('id', 0),
            'revision': None,
            'modified': metadata.get("modifiedDate", 0)
        }
        info.update(metadata)

        return info



# -*- coding: utf-8 -*-
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Command-line tools for assets."""

import argparse
import os
import warnings

import pkg_resources
from flask import current_app, json
from flask_assets import ManageAssets
from flask_script import Command, Option

from .registry import bundles


class AssetsCommand(ManageAssets):

    """Command-line operation for assets."""

    def run(self, args):
        """Run the command-line.

        It loads the bundles from the :py:data:`bundles registry
        <invenio_ext.assets.registry.bundles>`.

        """
        if not self.env:
            self.env = current_app.jinja_env.assets_environment
            self.log = current_app.logger

        for pkg, bundle in bundles:
            output = "{0}: {1.output}:".format(pkg, bundle)
            for content in bundle.contents:
                output += "\n - {0}".format(content)
            self.log.info(output)
            self.env.register(bundle.output, bundle)

        return super(AssetsCommand, self).run(args)


class BowerCommand(Command):

    """Command-line operation for bower."""

    option_list = (
        Option("-i", "--bower-json", help="base input file",
               dest="filename", default=None),
        Option("-o", "--output-file", help="write bower.json to output file",
               dest="output_file", metavar="FILENAME",
               type=argparse.FileType(mode='w')),
        Option("-x", "--override", help="(DEPRECATED) override the input file",
               dest="override", action="store_const", const=True),
    )

    def run(self, filename=None, override=False, output_file=None):
        """Generate a bower.json file.

        It comes with default values for the ignore. Name and version are set
        to be invenio's.
        """
        output = {
            "name": "invenio",
            "version": pkg_resources.get_distribution("invenio").version,
            "dependencies": {},
            "resolutions": {
                "jquery": "~1.11",  # <2 would pick 1.7.2 (thx jquery.treeview)
                "bootstrap": "~3.3"
            }
        }

        if filename and os.path.exists(filename):
            with open(filename, "r") as f:
                output = dict(output, **json.load(f))

        for pkg, bundle in bundles:
            if bundle.bower:
                current_app.logger.debug((pkg, bundle.bower))
            output['dependencies'].update(bundle.bower)

        # Remove together with override kwarg, and Option object.
        if override:
            warnings.warn("Use of --override is deprecated and will "
                          "be removed. Please use --output-file instead.",
                          DeprecationWarning)
            if filename and os.path.exists(filename) and output_file is None:
                output_file = open(filename, 'w')

        options = dict(indent=4)
        if output_file is None:
            print(json.dumps(output, **options)).encode("utf-8")
        else:
            json.dump(output, output_file, **options)
            output_file.close()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Admin."""

from __future__ import unicode_literals

from flask import abort, current_app
from flask_admin import AdminIndexView as FlaskAdminIndexView
from flask_admin import BaseView as FlaskBaseView
from flask_admin.contrib.sqla import ModelView as FlaskModelView
from flask_login import current_user

from invenio_ext.principal import permission_required
from invenio_ext.sslify import ssl_required


def can_acc_action(action):
    """Return getter/setter for can_<action> properties."""

    def fget(self):
        return self.acc_authorize(action)

    def fset(self, value):
        setattr(self, '_can_%s' % action, value)

    def fdel(self):
        delattr(self, '_can_%s' % action)

    return (fget, fset, fdel)


#
# Base classes
#
class BaseView(FlaskBaseView):

    """BaseView for administration interfaces."""

    acc_view_action = None

    can_view = property(*can_acc_action('view'))

    def acc_authorize(self, action):
        """Check authorization for a given action."""
        # First check if _can_<action> is set.
        if not getattr(self, '_can_%s' % action, True):
            return False

        # Next check if user is authorized to edit
        action = getattr(self, 'acc_%s_action' % action, None)
        if action:
            return permission_required(action)
        else:
            return current_user.is_super_admin

    def _handle_view(self, name, **kwargs):
        """The method will be executed before calling any view method."""
        if not current_user.is_authenticated():
            return current_app.login_manager.unauthorized()

        if not self.is_accessible():
            return abort(403)

    def is_accessible(self):
        """Check if admin interface is accessible by the current user."""
        if not current_user.is_authenticated():
            return False
        return self.can_view

    def create_blueprint(self, admin):
        """Ensure admin is only available over SSL."""
        self.blueprint = ssl_required(
            super(BaseView, self).create_blueprint(admin)
        )
        return self.blueprint


class ModelView(FlaskModelView, BaseView):

    """Invenio Admin base view for SQL alchemy models."""

    acc_edit_action = None
    acc_delete_action = None
    acc_create_action = None

    can_delete = property(*can_acc_action('delete'))
    can_edit = property(*can_acc_action('edit'))
    can_create = property(*can_acc_action('create'))


class AdminIndexView(FlaskAdminIndexView, BaseView):

    """Invenio admin index view that ensures Blueprint is being used."""

    # Ensures that templates and static files can be found
    import_name = 'flask_admin'



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2005, 2006, 2007, 2008, 2010, 2011, 2013, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""
    invenio_ext.email.errors
    ------------------------

    Contains standard error messages for email extension.
"""


class EmailError(Exception):
    """A generic email error."""

    def __init__(self, message):
        """Initialisation."""
        self.message = message

    def __str__(self):
        """String representation."""
        return repr(self.message)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Tests for Sentry sensitive data sanitation."""

from mock import Mock

from invenio_ext.logging.backends.sentry import InvenioSanitizeProcessor
from invenio_testing import InvenioTestCase

VARS = {
    'foo': 'bar',
    'password': 'hello',
    'access_token': 'hello',
}


class InvenioSantizeProcessorTest(InvenioTestCase):

    def _check_vars_sanitized(self, vars, proc):
        """Helper to check that keys have been sanitized."""
        self.assertTrue('foo' in vars)
        self.assertEquals(vars['foo'], 'bar')
        # Raven default processor takes care of this one
        self.assertTrue('password' in vars)
        self.assertEquals(vars['password'], 'hello')
        self.assertTrue('access_token' in vars)
        self.assertEquals(vars['access_token'], proc.MASK)

    def test_http(self):
        data = {
            'request': {
                'data': VARS,
                'env': VARS,
                'headers': VARS,
                'cookies': VARS,
            }
        }

        proc = InvenioSanitizeProcessor(Mock())
        result = proc.process(data)

        self.assertTrue('request' in result)
        http = result['request']
        for n in ('data', 'env', 'headers', 'cookies'):
            self.assertTrue(n in http)
            self._check_vars_sanitized(http[n], proc)

    def test_querystring_as_string(self):
        data = {
            'request': {
                'query_string': 'foo=bar&password=hello&access_token=hello',
            }
        }

        proc = InvenioSanitizeProcessor(Mock())
        result = proc.process(data)

        self.assertTrue('request' in result)
        http = result['request']
        self.assertEquals(
            http['query_string'],
            'foo=bar&password=hello&access_token=%(m)s' % dict(m=proc.MASK)
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Configuration of *Flask-Collect* extension."""

from flask_collect import Collect

collect = Collect()


def setup_app(app):
    """Set the application up with the correct static root."""
    def filter_(items):
        """Filter application blueprints."""
        order = [blueprint.name for blueprint in
                 app.extensions['registry']['blueprints']]

        def _key(item):
            if item.name in order:
                return order.index(item.name)
            return -1

        return sorted(items, key=_key)

    app.config.setdefault('COLLECT_FILTER', filter_)
    app.config.setdefault('COLLECT_STATIC_ROOT', app.static_folder)
    collect.init_app(app)

    # unsetting the static_folder so it's not picked up by collect.
    class FakeApp(object):
        name = "fakeapp"
        has_static_folder = False
        static_folder = None

    collect.app = FakeApp()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Unit test for the template extensions."""

from __future__ import absolute_import

from flask import url_for

from invenio_ext.template import render_template_to_string
from invenio_testing import InvenioTestCase, unittest


class TemplateTest(InvenioTestCase):
    """
    jinja2utils TestSuite.
    """

    def tplEqualToString(self, tpl, text, **ctx):
        self.assertEqual(
            render_template_to_string(tpl, _from_string=True, **ctx),
            text)

    def test_wrap_equal_to_prefix_and_suffix(self):
        wrap_tpl = '{{ test_variable|wrap(prefix="***", suffix="###") }}'
        pxsx_tpl = '{{ test_variable|prefix("***")|suffix("###") }}'
        # None is printed as empty string
        self.tplEqualToString(wrap_tpl, '', test_variable=None)
        self.tplEqualToString(pxsx_tpl, '', test_variable=None)
        # Nothing is appended to empty string
        self.tplEqualToString(wrap_tpl, '', test_variable='')
        self.tplEqualToString(pxsx_tpl, '', test_variable='')
        # x|prefix|suffix is equal to x|wrap
        self.tplEqualToString(wrap_tpl, '***test###', test_variable='test')
        self.tplEqualToString(pxsx_tpl, '***test###', test_variable='test')


class TemplateLoaderCase(InvenioTestCase):

    @property
    def config(self):
        cfg = super(TemplateLoaderCase, self).config
        cfg['PACKAGES'] = [
            'apps.first',
            'invenio_base',
            'apps.last',
        ]
        return cfg

    def test_fisrt_blueprint(self):
        response = self.client.get('/')
        self.assertEqual(response.data.strip(), 'First')
        self.assertNotEqual(response.data.strip(), 'Last')


class TemplateArgsTest(InvenioTestCase):

    """Test ``template_args`` decorator."""

    @classmethod
    def setup_app(cls, app):
        """Custom setup function."""
        from invenio_ext.template.context_processor import template_args
        from invenio_collections.views.collections import index

        @template_args(index)
        def foo():
            return {'foo': 'foo', 'baz': 'baz'}

        @template_args('collections.index', app=app)
        def bar():
            return {'bar': 'bar', 'baz': 'BAZ'}

    @property
    def config(self):
        from invenio_base.config import EXTENSIONS
        cfg = super(TemplateArgsTest, self).config
        cfg['EXTENSIONS'] = EXTENSIONS + [
            'test_ext_template.TemplateArgsTest',
        ]
        return cfg

    def test_template_args_loading(self):
        self.client.get(url_for('collections.index'))
        self.assertEqual(self.get_context_variable('foo'), 'foo')
        self.assertEqual(self.get_context_variable('bar'), 'bar')
        self.assertEqual(self.get_context_variable('baz'), 'BAZ')


class TemplateArgsLoadingTest(unittest.TestCase):

    """Test ``template_args`` decorator outside app context."""

    def test_broken_loading(self):
        from invenio_ext.template.context_processor import template_args

        def foo():
            return {'foo': 'foo'}

        self.assertRaises(Exception,
                          lambda: template_args('collections.index')(foo))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Registry for Jasmine spec files."""

import os
import re

from flask_registry import RegistryProxy
from werkzeug.utils import import_string

from invenio_ext.registry import DictModuleAutoDiscoverySubRegistry


class JasmineSpecsAutoDiscoveryRegistry(DictModuleAutoDiscoverySubRegistry):

    """Registry for Jasmine spec files.

    Looks into /testsuite/js/*.spec.js in each module.
    """

    pattern = re.compile("(?:.+\.js$)|(?:.+\.html$)")
    specs_folder = 'js'

    def __init__(self, *args, **kwargs):
        """Initialize registry."""
        super(JasmineSpecsAutoDiscoveryRegistry, self).__init__(
            'testsuite', **kwargs
        )

    def keygetter(self, key, original_value, new_value):
        """No key mapping."""
        return key

    def _walk_dir(self, pkg, base, root):
        """Recursively register *.spec.js/*.js files."""
        for root, dirs, files in os.walk(root):
            for name in files:
                if JasmineSpecsAutoDiscoveryRegistry.pattern.match(name):
                    filename = os.path.join(root, name)
                    filepath = "{0}/{1}".format(
                        pkg,
                        filename[len(base) + 1:]
                    )
                    self.register(filename, key=filepath)

    def _discover_module(self, pkg):
        """Load list of files from resource directory."""
        import_str = pkg + '.' + self.module_name

        try:
            module = import_string(import_str, silent=self.silent)
            if module is not None:
                for p in module.__path__:
                    specsfolder = os.path.join(p, self.specs_folder)
                    if os.path.isdir(specsfolder):
                        self._walk_dir(pkg, specsfolder, specsfolder)
        except ImportError as e:  # pylint: disable=C0103
            self._handle_importerror(e, pkg, import_str)
        except SyntaxError as e:
            self._handle_syntaxerror(e, pkg, import_str)


specs = RegistryProxy("jasmine.specs", JasmineSpecsAutoDiscoveryRegistry)
